"""Unit tests for additions to django.contrib.auth.models.User."""

from __future__ import annotations

from typing import TYPE_CHECKING
from uuid import UUID, uuid4

import kgb
from django_assert_queries.testing import assert_queries
from django.contrib.auth.models import AnonymousUser, User
from django.db.models import Q
from djblets.testing.decorators import add_fixtures

from reviewboard.accounts.models import Profile
from reviewboard.site.models import LocalSite
from reviewboard.testing import TestCase

if TYPE_CHECKING:
    from django_assert_queries import ExpectedQueries


class UserTests(kgb.SpyAgency, TestCase):
    """Unit tests for additions to django.contrib.auth.models.User."""

    fixtures = ['test_users']

    def test_get_profile(self):
        """Testing User.get_profile"""
        user = self.create_user(username='test1')
        profile = Profile.objects.create(user=user)

        # 1 query:
        #
        # 1. Fetch profile
        queries = [
            {
                'model': Profile,
                'where': Q(user=user),
            },
        ]

        with assert_queries(queries):
            new_profile = user.get_profile()

        self.assertEqual(new_profile, profile)
        self.assertIsNot(new_profile, profile)
        self.assertIs(new_profile.user, user)

        # A second call should hit cache.
        with self.assertNumQueries(0):
            self.assertIs(user.get_profile(), new_profile)

    def test_get_profile_with_no_profile(self):
        """Testing User.get_profile with no existing profile"""
        user = self.create_user(username='test1')

        # 4 queries:
        #
        # 1. Attempt to fetch profile
        # 2. Create savepoint
        # 3. Create Profile
        # 4. Release savepoint
        queries = [
            {
                'model': Profile,
                'where': Q(user=user),
            },
            {
                'model': Profile,
                'type': 'INSERT',
            },
        ]

        with assert_queries(queries, num_statements=4):
            new_profile = user.get_profile()

        self.assertIs(new_profile.user, user)
        self.assertIsNotNone(new_profile.pk)

        # A second call should hit cache.
        with self.assertNumQueries(0):
            self.assertIs(user.get_profile(), new_profile)

    def test_get_profile_with_prefetch_related(self):
        """Testing User.get_profile with prefetch_related"""
        user = self.create_user(username='test1')
        Profile.objects.create(user=user)

        # Now re-fetch.
        #
        # 2 queries:
        #
        # 1. Fetch users
        # 2. Fetch profiles for fetched user IDs
        queries = [
            {
                'model': User,
                'where': Q(username='test1'),
            },
            {
                'model': Profile,
                'where': Q(user__in=[user]),
            },
        ]

        with assert_queries(queries):
            user = list(
                User.objects
                .filter(username='test1')
                .prefetch_related('profile_set')
            )[0]

        with self.assertNumQueries(0):
            profile = user.get_profile()

        self.assertIs(profile.user, user)
        self.assertIsNotNone(profile.pk)

    def test_get_profile_with_select_related(self):
        """Testing User.get_profile with select_related"""
        user = self.create_user(username='test1')
        Profile.objects.create(user=user)

        # Now re-fetch.
        #
        # 1 query:
        #
        # 1. Fetch users + profiles
        queries = [
            {
                'model': User,
                'select_related': {'profile'},
                'where': Q(username='test1'),
            },
        ]

        with assert_queries(queries):
            user = list(
                User.objects
                .filter(username='test1')
                .select_related('profile')
            )[0]

        with self.assertNumQueries(0):
            profile = user.get_profile()

        self.assertIs(profile.user, user)
        self.assertIsNotNone(profile.pk)

    def test_get_profile_with_select_related_none_create_if_missing_true(
        self,
    ) -> None:
        """Testing User.get_profile with select_related with no profile
        in database and create_if_missing=True
        """
        user = self.create_user(username='test1')

        equeries: ExpectedQueries = [
            {
                'model': User,
                'select_related': {'profile'},
                'where': Q(username='test1'),
            },
        ]

        with assert_queries(equeries):
            user = list(
                User.objects
                .filter(username='test1')
                .select_related('profile')
            )[0]

        equeries = [
            {
                'model': Profile,
                'where': Q(user=user),
            },
            {
                'model': Profile,
                'type': 'INSERT',
            },
        ]

        with assert_queries(equeries):
            profile = user.get_profile(create_if_missing=True)

        self.assertIsNotNone(profile)
        self.assertIsNotNone(profile.pk)

    def test_get_profile_with_select_related_none_create_if_missing_false(
        self,
    ) -> None:
        """Testing User.get_profile with select_related with no profile
        in database and create_if_missing=False
        """
        user = self.create_user(username='test1')

        equeries: ExpectedQueries = [
            {
                'model': User,
                'select_related': {'profile'},
                'where': Q(username='test1'),
            },
        ]

        with assert_queries(equeries):
            user = list(
                User.objects
                .filter(username='test1')
                .select_related('profile')
            )[0]

        with self.assertNumQueries(0):
            with self.assertRaises(Profile.DoesNotExist):
                user.get_profile(create_if_missing=False)

    def test_get_profile_with_no_profile_and_create_if_missing_false(self):
        """Testing User.get_profile with no existing profile and
        create_if_missing=False
        """
        user = self.create_user(username='test1')

        # 1 query:
        #
        # 1. Attempt to fetch profile
        queries = [
            {
                'model': Profile,
                'where': Q(user=user),
            },
        ]

        with assert_queries(queries):
            with self.assertRaises(Profile.DoesNotExist):
                user.get_profile(create_if_missing=False)

    def test_get_profile_with_cached_only_and_in_cache(self):
        """Testing User.get_profile with cached_only=True and profile already
        in object cache
        """
        user = self.create_user(username='test1')
        profile = user.get_profile()

        with self.assertNumQueries(0):
            new_profile = user.get_profile(cached_only=True)

        self.assertIs(new_profile, profile)

    def test_get_profile_with_cached_only_and_not_in_cache(self):
        """Testing User.get_profile with cached_only=True and profile not
        in object cache
        """
        user = self.create_user(username='test1')
        Profile.objects.create(user=user)

        with self.assertNumQueries(0):
            self.assertIsNone(user.get_profile(cached_only=True))

    def test_get_profile_with_return_is_new_and_new(self):
        """Testing User.get_profile with return_is_new=True and profile is new
        """
        user = self.create_user(username='test1')
        result = user.get_profile(return_is_new=True)

        self.assertIsInstance(result, tuple)
        self.assertEqual(len(result), 2)
        self.assertIsInstance(result[0], Profile)
        self.assertTrue(result[1])

    def test_get_profile_with_return_is_new_and_not_new(self):
        """Testing User.get_profile with return_is_new=True and profile is new
        """
        user = self.create_user(username='test1')
        Profile.objects.create(user=user)

        result = user.get_profile(return_is_new=True)

        self.assertIsInstance(result, tuple)
        self.assertEqual(len(result), 2)
        self.assertIsInstance(result[0], Profile)
        self.assertFalse(result[1])

    def test_is_profile_visible_with_public(self):
        """Testing User.is_profile_visible with public profiles"""
        user1 = User.objects.get(username='admin')
        user2 = User.objects.get(username='doc')

        self.assertTrue(user1.is_profile_visible(user2))

    def test_is_profile_visible_with_private(self):
        """Testing User.is_profile_visible with private profiles"""
        user1 = User.objects.get(username='admin')
        user2 = User.objects.get(username='doc')

        profile = user1.get_profile()
        profile.is_private = True
        profile.save(update_fields=('is_private',))

        self.assertFalse(user1.is_profile_visible(user2))
        self.assertTrue(user1.is_profile_visible(user1))

        user2.is_staff = True
        self.assertTrue(user1.is_profile_visible(user2))

    def test_is_profile_visible_unauthenticated(self):
        """Testing User.is_profile_visible with an unauthenticated user"""
        user = User.objects.get(username='doc')

        self.assertFalse(user.is_profile_visible(AnonymousUser()))

    def test_is_profile_visible_no_user(self):
        """Testing User.is_profile_visible with no user"""
        user = User.objects.get(username='doc')

        self.assertFalse(user.is_profile_visible(None))

    def test_is_profile_visible_staff(self):
        """Testing User.is_profile_public with a staff user"""
        user = User.objects.get(username='doc')
        admin = User.objects.get(username='admin')

        profile = user.get_profile()
        profile.is_private = True
        profile.save(update_fields=('is_private',))

        self.assertTrue(user.is_profile_visible(admin))

    def test_is_profile_visible_owner(self):
        """Testing User.is_profile_visible for the profile owner"""
        user = User.objects.get(username='doc')
        profile = user.get_profile()
        profile.is_private = True
        profile.save(update_fields=('is_private',))

        self.assertTrue(user.is_profile_visible(user))

    def test_is_profile_visible_local_site_member(self):
        """Testing User.is_profile_visible for a LocalSite member viewing a
        LocalSite member with a public profile
        """
        to_view = User.objects.get(username='doc')
        viewer = User.objects.get(username='grumpy')

        site = LocalSite.objects.create()
        site.users.add(to_view, viewer)

        self.assertTrue(to_view.is_profile_visible(viewer))

    def test_is_profile_visible_local_site_member_private(self):
        """Testing User.is_profile_visible for a LocalSite member viewing a
        LocalSite member with a private profile
        """
        to_view = User.objects.get(username='doc')
        viewer = User.objects.get(username='grumpy')

        profile = to_view.get_profile()
        profile.is_private = True
        profile.save(update_fields=('is_private',))

        site = LocalSite.objects.create()
        site.users.add(to_view, viewer)

        self.assertFalse(to_view.is_profile_visible(viewer))

    def test_is_profile_visible_local_site_admin(self):
        """Testing user.is_profile_visible for a LocalSite admin viewing a
        LocalSite member with a public profile
        """
        to_view = User.objects.get(username='doc')
        viewer = User.objects.get(username='grumpy')

        site = LocalSite.objects.create()
        site.users.add(to_view, viewer)
        site.admins.add(viewer)

        self.assertTrue(to_view.is_profile_visible(viewer))

    def test_is_profile_visible_local_site_admin_private(self):
        """Testing user.is_profile_visible for a LocalSite admin viewing a
        LocalSite member with a private profile
        """
        to_view = User.objects.get(username='doc')
        viewer = User.objects.get(username='grumpy')

        profile = to_view.get_profile()
        profile.is_private = True
        profile.save(update_fields=('is_private',))

        site = LocalSite.objects.create()
        site.users.add(to_view, viewer)
        site.admins.add(viewer)

        self.assertTrue(to_view.is_profile_visible(viewer))

    def test_is_admin_for_user_admin_vs_user(self):
        """Testing User.is_admin_for_user for an admin"""
        admin = User.objects.get(username='admin')
        user = User.objects.get(username='doc')

        with self.assertNumQueries(0):
            self.assertTrue(admin.is_admin_for_user(user))

    def test_is_admin_for_user_admin_vs_none(self):
        """Testing User.is_admin_for_user for an admin when the user is None"""
        admin = User.objects.get(username='admin')

        with self.assertNumQueries(0):
            self.assertTrue(admin.is_admin_for_user(None))

    def test_is_admin_for_user_admin_vs_anonymous(self):
        """Testing User.is_admin_for_user for an admin when the user is
        anonymous
        """
        admin = User.objects.get(username='admin')

        with self.assertNumQueries(0):
            self.assertTrue(admin.is_admin_for_user(AnonymousUser()))

    def test_is_admin_for_user_user_vs_user(self):
        """Testing User.is_admin_for_user for a regular user"""
        user = User.objects.get(username='doc')

        # This has the side-effect of pre-fetching stats, so they don't
        # interfere with query counts below.
        self.assertFalse(LocalSite.objects.has_local_sites())

        with self.assertNumQueries(0):
            self.assertFalse(user.is_admin_for_user(user))

    def test_is_admin_for_user_localsite_admin_vs_localsite_user(self):
        """Testing User.is_admin_for_user for a LocalSite admin when the user
        is a member of that LocalSite
        """
        site_admin = self.create_user(username='site_admin')
        site_user1 = self.create_user(username='user1')
        site_user2 = self.create_user(username='user2')

        local_site = LocalSite.objects.create()
        local_site.admins.add(site_admin)
        local_site.users.add(site_admin, site_user1, site_user2)

        # This has the side-effect of pre-fetching stats, so they don't
        # interfere with query counts below.
        self.assertTrue(LocalSite.objects.has_local_sites())

        # 4 queries (all from get_local_site_stats() calls):
        #
        # 1. Fetch LocalSite IDs for site_admin user membership.
        # 2. Fetch LocalSite IDs for site_admin admin membership.
        # 3. Fetch LocalSite IDs for site_user1 user membership.
        # 4. Fetch LocalSite IDs for site_user1 admin membership.
        queries = [
            {
                'model': LocalSite.users.through,
                'values_select': ('localsite_id',),
                'where': Q(user=site_admin),
            },
            {
                'model': LocalSite.admins.through,
                'values_select': ('localsite_id',),
                'where': Q(user=site_admin),
            },
            {
                'model': LocalSite.users.through,
                'values_select': ('localsite_id',),
                'where': Q(user=site_user1),
            },
            {
                'model': LocalSite.admins.through,
                'values_select': ('localsite_id',),
                'where': Q(user=site_user1),
            },
        ]

        with assert_queries(queries):
            self.assertTrue(site_admin.is_admin_for_user(site_user1))

        # A second call should reuse cached stats.
        with self.assertNumQueries(0):
            self.assertTrue(site_admin.is_admin_for_user(site_user1))

        # For another user, get_local_site_stats() will be called again, but
        # only for that user.
        #
        # 2 queries (all from get_local_site_stats() calls):
        #
        # 1. Fetch LocalSite IDs for site_user2 user membership.
        # 2. Fetch LocalSite IDs for site_user2 admin membership.
        queries = [
            {
                'model': LocalSite.users.through,
                'values_select': ('localsite_id',),
                'where': Q(user=site_user2),
            },
            {
                'model': LocalSite.admins.through,
                'values_select': ('localsite_id',),
                'where': Q(user=site_user2),
            },
        ]

        with assert_queries(queries):
            self.assertTrue(site_admin.is_admin_for_user(site_user2))

    @add_fixtures(['test_site'])
    def test_is_admin_for_user_localsite_admin_vs_other_localsite_user(self):
        """Testing User.is_admin_for_user for a LocalSite admin when the user
        is a member of another LocalSite
        """
        site1_admin = self.create_user(username='site1_admin')
        site1_user1 = self.create_user(username='site1_user1')
        site2_user1 = self.create_user(username='site2_user1')
        site2_user2 = self.create_user(username='site2_user2')

        local_site1 = LocalSite.objects.create(name='site1')
        local_site1.admins.add(site1_admin)
        local_site1.users.add(site1_admin, site1_user1)

        local_site2 = LocalSite.objects.create(name='site2')
        local_site2.users.add(site2_user1, site2_user2)

        # This has the side-effect of pre-fetching stats, so they don't
        # interfere with query counts below.
        self.assertTrue(LocalSite.objects.has_local_sites())

        # 4 queries (all from get_local_site_stats() calls):
        #
        # 1. Fetch LocalSite IDs for site1_admin user membership.
        # 2. Fetch LocalSite IDs for site1_admin admin membership.
        # 3. Fetch LocalSite IDs for site2_user1 user membership.
        # 4. Fetch LocalSite IDs for site2_user1 admin membership.
        queries = [
            {
                'model': LocalSite.users.through,
                'values_select': ('localsite_id',),
                'where': Q(user=site1_admin),
            },
            {
                'model': LocalSite.admins.through,
                'values_select': ('localsite_id',),
                'where': Q(user=site1_admin),
            },
            {
                'model': LocalSite.users.through,
                'values_select': ('localsite_id',),
                'where': Q(user=site2_user1),
            },
            {
                'model': LocalSite.admins.through,
                'values_select': ('localsite_id',),
                'where': Q(user=site2_user1),
            },
        ]

        with assert_queries(queries):
            self.assertFalse(site1_admin.is_admin_for_user(site2_user1))

        # A second call should reuse cached stats.
        with self.assertNumQueries(0):
            self.assertFalse(site1_admin.is_admin_for_user(site2_user1))

        # For another user, get_local_site_stats() will be called again, but
        # only for that user.
        #
        # 2 queries (all from get_local_site_stats() calls):
        #
        # 1. Fetch LocalSite IDs for site2_user2 user membership.
        # 2. Fetch LocalSite IDs for site2_user2 admin membership.
        queries = [
            {
                'model': LocalSite.users.through,
                'values_select': ('localsite_id',),
                'where': Q(user=site2_user2),
            },
            {
                'model': LocalSite.admins.through,
                'values_select': ('localsite_id',),
                'where': Q(user=site2_user2),
            },
        ]

        with assert_queries(queries):
            self.assertFalse(site1_admin.is_admin_for_user(site2_user2))

    def test_get_local_site_stats(self):
        """Testing User.get_local_site_stats"""
        self.spy_on(uuid4, op=kgb.SpyOpReturnInOrder([
            # First will be for LocalSite.objects.get_stats().
            UUID('00000000-0000-0000-0000-000000000001'),

            # Second will be for User.get_local_site_stats().
            UUID('00000000-0000-0000-0000-000000000002'),
        ]))

        user = self.create_user(username='site1_admin')

        local_site1 = LocalSite.objects.create(name='site1')
        local_site1.admins.add(user)
        local_site1.users.add(user)

        local_site2 = LocalSite.objects.create(name='site2')
        local_site2.users.add(user)

        local_site3 = LocalSite.objects.create(name='site3')
        local_site3.admins.add(user)

        # Pre-fetch the global LocalSite stats, so it doesn't impact the
        # query count below.
        LocalSite.objects.get_stats()

        # 2 queries:
        #
        # 1. User LocalSite membership count
        # 2. User LocalSite admined count
        queries = [
            {
                'model': LocalSite.users.through,
                'values_select': ('localsite_id',),
                'where': Q(user=user),
            },
            {
                'model': LocalSite.admins.through,
                'values_select': ('localsite_id',),
                'where': Q(user=user),
            },
        ]

        with assert_queries(queries):
            self.assertEqual(
                user.get_local_site_stats(),
                {
                    'admined_local_site_ids': [local_site1.pk, local_site3.pk],
                    'local_site_ids': [local_site1.pk, local_site2.pk],
                    'state_uuid': '00000000-0000-0000-0000-000000000002',
                })

        # A second call should hit cache.
        with self.assertNumQueries(0):
            self.assertEqual(
                user.get_local_site_stats(),
                {
                    'admined_local_site_ids': [local_site1.pk, local_site3.pk],
                    'local_site_ids': [local_site1.pk, local_site2.pk],
                    'state_uuid': '00000000-0000-0000-0000-000000000002',
                })

    def test_get_local_site_stats_after_state_uuid_change(self):
        """Testing User.get_local_site_stats after
        LocalSite.objects.get_stats() state_uuid change
        """
        self.spy_on(uuid4, op=kgb.SpyOpReturnInOrder([
            # First will be for LocalSite.objects.get_stats().
            UUID('00000000-0000-0000-0000-000000000001'),

            # Second will be for User.get_local_site_stats().
            UUID('00000000-0000-0000-0000-000000000002'),

            # Third will be LocalSite.objects.get_stats() again.
            UUID('00000000-0000-0000-0000-000000000003'),

            # Fourth will be for User.get_local_site_stats() again.
            UUID('00000000-0000-0000-0000-000000000004'),
        ]))

        user = self.create_user(username='site1_admin')

        local_site1 = LocalSite.objects.create(name='site1')
        local_site1.admins.add(user)
        local_site1.users.add(user)

        local_site2 = LocalSite.objects.create(name='site2')
        local_site2.users.add(user)

        local_site3 = LocalSite.objects.create(name='site3')
        local_site3.admins.add(user)

        # Pre-fetch the global LocalSite stats, so it doesn't impact the
        # query count below.
        LocalSite.objects.get_stats()

        # 2 queries:
        #
        # 1. User LocalSite membership count
        # 2. User LocalSite admined count
        queries = [
            {
                'model': LocalSite.users.through,
                'values_select': ('localsite_id',),
                'where': Q(user=user),
            },
            {
                'model': LocalSite.admins.through,
                'values_select': ('localsite_id',),
                'where': Q(user=user),
            },
        ]

        with assert_queries(queries):
            self.assertEqual(
                user.get_local_site_stats(),
                {
                    'admined_local_site_ids': [local_site1.pk, local_site3.pk],
                    'local_site_ids': [local_site1.pk, local_site2.pk],
                    'state_uuid': '00000000-0000-0000-0000-000000000002',
                })

        # This should impact LocalSite stats.
        self.create_local_site(name='site4')
        LocalSite.objects.get_stats()

        # A second call should not hit cache.
        with assert_queries(queries):
            self.assertEqual(
                user.get_local_site_stats(),
                {
                    'admined_local_site_ids': [local_site1.pk, local_site3.pk],
                    'local_site_ids': [local_site1.pk, local_site2.pk],
                    'state_uuid': '00000000-0000-0000-0000-000000000004',
                })

    def test_get_local_site_stats_with_no_local_sites(self):
        """Testing User.get_local_site_stats with no LocalSites in database"""
        self.spy_on(uuid4, op=kgb.SpyOpReturnInOrder([
            # This will be for LocalSite.objects.get_stats().
            UUID('00000000-0000-0000-0000-000000000001'),
        ]))

        user = self.create_user(username='site1_admin')

        # Pre-fetch the global LocalSite stats, so it doesn't impact the
        # query count below.
        LocalSite.objects.get_stats()

        with self.assertNumQueries(0):
            self.assertEqual(
                user.get_local_site_stats(),
                {
                    'admined_local_site_ids': [],
                    'local_site_ids': [],
                    'state_uuid': '00000000-0000-0000-0000-000000000001',
                })



"""Common utility functions for working with Subversion."""

from __future__ import annotations

import re
from typing import List


AUTHOR_KEYWORDS = [b'author', b'lastchangedby']
DATE_KEYWORDS = [b'date', b'lastchangeddate']
REVISION_KEYWORDS = [b'revision', b'lastchangedrevision', b'rev']
URL_KEYWORDS = [b'headurl', b'url']
ID_KEYWORDS = [b'id']
HEADER_KEYWORDS = [b'header']


# Mapping of keywords to known aliases
keyword_aliases = {
    # Standard keywords
    b'author': AUTHOR_KEYWORDS,
    b'date': DATE_KEYWORDS,
    b'header': HEADER_KEYWORDS,
    b'headurl': URL_KEYWORDS,
    b'id': ID_KEYWORDS,
    b'revision': REVISION_KEYWORDS,

    # Aliases
    b'lastchangedby': AUTHOR_KEYWORDS,
    b'lastchangeddate': DATE_KEYWORDS,
    b'lastchangedrevision': REVISION_KEYWORDS,
    b'rev': REVISION_KEYWORDS,
    b'url': URL_KEYWORDS,
}


def has_expanded_svn_keywords(
    data: bytes,
) -> bool:
    """Return whether file data appears to have expanded SVN keywords.

    This can be used to determine whether it's worth fetching information from
    a Subversion repository to expand file data.

    Args:
        data (bytes):
            The file content.

    Returns:
        bool:
        ``True`` if the data appears to have SVN keywords. ``False`` if it
        does not.
    """
    assert isinstance(data, bytes), (
        'data must be a byte string, not %s' % type(data))

    return bool(re.search(br'\$[A-Za-z0-9]+::?\s+([^\$\s]+)[^\$]*\$',
                          data))


def collapse_svn_keywords(
    data: bytes,
    keyword_str: bytes,
) -> bytes:
    """Collapse SVN keywords in string.

    SVN allows for several keywords (such as ``$Id$`` and ``$Revision$``) to be
    expanded, though these keywords are limited to a fixed set (and associated
    aliases) and must be enabled per-file.

    Keywords can take two forms: ``$Keyword$`` and ``$Keyword::     $``.
    The latter allows the field to take a fixed size when expanded.

    When we cat a file on SVN, the keywords come back expanded, which isn't
    good for us as we need to diff against the collapsed version. This
    function makes that transformation.

    Args:
        data (bytes):
            The file content.

        keyword_str (bytes):
            One or more keywords, separated by spaces.

    Returns:
        bytes:
        The file content with keywords collapsed.
    """
    def repl(m):
        if m.group(2):
            return b'$%s::%s$' % (m.group(1), b' ' * len(m.group(3)))

        return b'$%s$' % m.group(1)

    assert isinstance(data, bytes), (
        'data must be a byte string, not %s' % type(data))
    assert isinstance(keyword_str, bytes), (
        'keyword_str must be a byte string, not %s' % type(keyword_str))

    # Get any aliased keywords
    keywords: List[bytes] = [
        re.escape(keyword)
        for name in re.split(br'\W+', keyword_str)
        for keyword in keyword_aliases.get(name.lower(), [])
    ]

    return re.sub(br'\$(%s):(:?)([^\$\n\r]*)\$' % b'|'.join(keywords),
                  repl, data, flags=re.IGNORECASE)



"""Unit tests for reviewboard.admin.forms.change_form.ChangeFormFieldset."""

from django.contrib.admin.helpers import AdminForm
from django.contrib.auth.models import User

from reviewboard.admin import admin_site
from reviewboard.admin.forms.change_form import (ChangeFormFieldset,
                                                 ChangeFormRow)
from reviewboard.testing.testcase import TestCase
from reviewboard.admin.templatetags.rbadmintags import change_form_fieldsets


class ChangeFormFieldsetTests(TestCase):
    """Unit tests for ChangeFormFieldset."""

    def setUp(self):
        super(ChangeFormFieldsetTests, self).setUp()

        request = self.create_http_request()
        model_admin = admin_site.get_model_admin(User)

        self.admin_form = AdminForm(
            form=model_admin.get_form(request)(),
            fieldsets=list(model_admin.get_fieldsets(request)),
            prepopulated_fields=model_admin.get_prepopulated_fields(request),
            readonly_fields=model_admin.get_readonly_fields(request),
            model_admin=model_admin)

    def test_init_normalizes_classes(self):
        """Testing ChangeFormFieldset.__init__ normalizes Django fieldset CSS
        classes
        """
        fieldset = ChangeFormFieldset(form=self.admin_form,
                                      classes=('collapse', 'wide', '-is-test'))
        self.assertEqual(
            fieldset.classes,
            'rb-c-form-fieldset -can-collapse -is-collapsed -is-wide -is-test')

    def test_collapsed_with_true(self):
        """Testing ChangeFormFieldset.collapsed with Django collapse CSS class
        """
        fieldset = ChangeFormFieldset(form=self.admin_form,
                                      classes=('collapse',))
        self.assertTrue(fieldset.collapsed)

    def test_collapsed_with_false(self):
        """Testing ChangeFormFieldset.collapsed without Django collapse CSS
        class
        """
        fieldset = ChangeFormFieldset(form=self.admin_form)
        self.assertFalse(fieldset.collapsed)

    def test_iter(self):
        """Testing ChangeFormFieldset.__iter__"""
        fieldset = list(change_form_fieldsets(self.admin_form))[0]

        rows = list(fieldset)
        self.assertGreater(len(rows), 0)

        for row in rows:
            self.assertIsInstance(row, ChangeFormRow)



#!/usr/bin/env python3
"""Builder for the code safety confusables dataset module.

This will pull down the latest lists of Unicode categories and confusables
from unicode.org and generate a mapping of confusables that we want to scan
for when displaying diffs. The result will be a new
:file:`reviewboard/codesafety/_unicode_confusables.py` file.

Version Changed:
    5.0.2:
    Updated to generate mappings of languages (aliases) to code ranges, and
    lists of found Unicode aliases for customization.

Version Added:
    5.0
"""

from __future__ import annotations

import os
import re
import sys
from typing import Any, Dict, Iterator, List, Optional, Tuple
from urllib.request import urlopen

if sys.version_info[:2] >= (3, 9):
    # Python >= 3.9 keeps Pattern[...] in re.
    from re import Pattern
else:
    # Python <= 3.8 keeps Pattern[...] in typing.
    from typing import Pattern

from typing_extensions import TypeAlias


ConfusableEntry: TypeAlias = Tuple[int, str, str]


CATEGORIES_URL = 'ftp://ftp.unicode.org/Public/UNIDATA/Scripts.txt'
CONFUSABLES_URL = \
    'ftp://ftp.unicode.org/Public/security/latest/confusables.txt'


#: A list of Unicode aliases we'll display for customization.
#:
#: Each of these can be enabled or disabled for Unicode Confusable checks by
#: administrators.
#:
#: Version Added:
#:     5.0.2
DISPLAY_ALIASES = {
    'ARABIC',
    'ARMENIAN',
    'BENGALI',
    'CYRILLIC',
    'DEVANAGARI',
    'GREEK',
    'GUJARATI',
    'GURMUKHI',
    'KANNADA',
    'KATAKANA',
    'ORIYA',
    'TELUGU',
}

scripts_dir = os.path.abspath(os.path.dirname(__file__))
dest_filename = os.path.abspath(os.path.join(
    scripts_dir, '..', '..', 'reviewboard', 'codesafety', '_confusables.py'))


categories_data: Dict[str, Any] = {}
aliases: Dict[str, List[str]] = {}


def _make_codepoints_key_path(
    codepoint: int,
) -> Tuple[int, int, int, int]:
    """Return a key path used for a Unicode codepoint.

    Args:
        codepoint (int):
            The Unicode codepoint to generate the key for.

    Returns:
        tuple:
        The key path.
    """
    return (
        codepoint & ~0xFFFF,
        codepoint & ~0xFFF,
        codepoint & ~0xFF,
        codepoint & ~0xF,
    )


def get_alias(
    codepoint: int,
) -> Tuple[Optional[str], Optional[str]]:
    """Return the category alias for a codepoint.

    Version Changed:
        5.0.2:
        Updated to return a tuple of results.

    Args:
        codepoint (int):
            The codepoint.

    Returns:
        tuple:
        A 2-tuple of:

        Tuple:
            0 (str):
                The normalized alias ID.

            1 (str):
                The human-readable alias name.

        These will both be ``None`` if an alias was not found.
    """
    key = _make_codepoints_key_path(codepoint)

    try:
        codepoint_ranges = \
            categories_data['codepoints'][key[0]][key[1]][key[2]][key[3]]
    except KeyError:
        return None, None

    for codepoint_range in codepoint_ranges:
        if codepoint_range[0] <= codepoint <= codepoint_range[1]:
            return categories_data['aliases'][codepoint_range[2]]

    return None, None


def _load_data(
    url: str,
    line_re: Pattern[str],
) -> Iterator[Dict]:
    """Download and iterate through a Unicode dataset.

    Args:
        url (str):
            The URL to download.

        line_re (re.RegexObject):
            The regex used to match lines.

    Yields:
        dict:
        Matched groups for each valid match of the regex.
    """
    data = urlopen(url).read()
    lines = data.decode('utf-8-sig').split('\n')

    for line in lines:
        if not line or line.startswith('#'):
            continue

        m = line_re.match(line)
        assert m, repr(line)

        yield m.groupdict()


def build_categories() -> None:
    """Download and build data on Unicode categories.

    This will download the Unicode categories dataset and parse it. It will
    populate :py:data:`categories_data` with lists of categories and aliases,
    and a tree of codepoint ranges.
    """
    # Note that we're acting like we're going to grab all categories from
    # the dataset, though in practice we're going to limit it to COMMON and
    # LATIN entries.
    #
    # This is because right now, we only need it for Unicode confusable maps,
    # but as we implement more Unicode checks, we will likely need to fetch
    # and store more information. Given the complexity of this, it's better to
    # have a more future-proof, "correct" implementation up-front.
    LINE_RE = re.compile(
        r'^(?P<codepoint_from>[0-9A-F]+)'
        r'(?:\.\.(?P<codepoint_through>[0-9A-F]+))?'
        r'\s*; (?P<alias>\w+) # (?P<category>[\w]+)',
        re.UNICODE)

    aliases: List[Tuple[str, str]] = []
    alias_id_map: Dict[str, int] = {}

    categories: List[str] = []
    category_id_map: Dict[str, int] = {}

    codepoint_ranges: Dict[int, Dict] = {}

    for info in _load_data(CATEGORIES_URL, LINE_RE):
        alias: str = info['alias']
        alias_id = alias.upper()

        if alias_id not in alias_id_map:
            alias_id_map[alias_id] = len(aliases)
            aliases.append((alias_id, alias.replace('_', ' ')))

        category = info['category']

        if category not in category_id_map:
            category_id_map[category] = len(categories)
            categories.append(category)

        codepoint_from_s = info['codepoint_from']
        codepoint_through_s = info['codepoint_through'] or codepoint_from_s

        codepoint_from = int(codepoint_from_s, 16)
        codepoint_through = int(codepoint_through_s, 16)

        # Split into subtables. Key off from some prefix.
        prev_key: Optional[Tuple[int, int, int, int]] = None
        cur_range: Optional[Tuple[int, int, int, int]] = None

        # We need a quick way to look up Unicode codepoints, but it's too
        # expensive to maintain a mapping of every codepoint. So, instead
        # we have a 5-level tree.
        #
        # The first 4 levels are increasingly specific masks of starting
        # codepoint ranges, with the 5th level being the codepoints in that
        # range.
        #
        # Codepoint ranges are split up as needed to fit in the correct range.
        #
        # As an example, if we were storing category range 1F400..1F6D7
        # (RAT..ELEVATOR):
        #
        #     10000: {
        #         1F000: {
        #             1F400: {
        #                 1F400: [1F400..1F409],
        #                 ...
        #                 1F490: [1F490..1F499],
        #             },
        #             ..
        #             1F600: {
        #                 1F600: [1F600..1F609],
        #                 ...
        #                 1F6D0: [1F6D0..1F6D7],
        #             }
        #         }
        #     }
        #
        # In practice, the leafs often have more than one range, particularly
        # for the lower codepoint ranges.
        #
        # This is easy to build and fast for lookup.
        for codepoint in range(codepoint_from, codepoint_through + 1):
            key = _make_codepoints_key_path(codepoint)

            if key != prev_key:
                if cur_range:
                    assert prev_key is not None

                    codepoints = (
                        codepoint_ranges
                        .setdefault(prev_key[0], {})
                        .setdefault(prev_key[1], {})
                        .setdefault(prev_key[2], {})
                        .setdefault(prev_key[3], [])
                    )
                    codepoints.append(cur_range)

                cur_range = (
                    codepoint,
                    codepoint,
                    alias_id_map[alias_id],
                    category_id_map[category],
                )
            else:
                assert cur_range is not None

                cur_range = (
                    cur_range[0],
                    codepoint,
                    cur_range[2],
                    cur_range[3],
                )

            prev_key = key

        if prev_key:
            codepoints = (
                codepoint_ranges
                .setdefault(prev_key[0], {})
                .setdefault(prev_key[1], {})
                .setdefault(prev_key[2], {})
                .setdefault(prev_key[3], [])
            )
            codepoints.append(cur_range)

    categories_data.update({
        'aliases': aliases,
        'categories': categories,
        'codepoints': codepoint_ranges,
    })


def update_confusables() -> List[ConfusableEntry]:
    """Download and build data on Unicode confusables/homoglyphs.

    This will download the Unicode confusables dataset and parse it. It will
    extract confusables that can be confused with COMMON or LATIN characters
    (skipping lower codepoints, to avoid issues with, say, "1" and "l"
    appearing in the data), and return a mapping of confusable characters
    to the COMMON/LATIN characters they may be confused with.

    Returns:
        list:
        A list of confusables. Each entry is a tuple in the form of:

        Tuple:
            0 (int):
                The codepoint of the confusable character.

            1 (str):
                The confusable character.

            2 (str):
                The character confused with.
    """
    LINE_RE = re.compile(
        r'^(?P<confusable_cp>[0-9A-F ]+) ;\t'
        r'(?P<confused_with_cp>[0-9A-F ]+) ;.*',
        re.UNICODE)

    confusables: List[ConfusableEntry] = []

    for info in _load_data(CONFUSABLES_URL, LINE_RE):
        # Parse and check the character that another may be confused with.
        confused_with_cp_strs = info['confused_with_cp'].split(' ')

        if len(confused_with_cp_strs) != 1:
            # We don't care about any confusables with multi-codepoint
            # characters.
            continue

        confused_with_cp = int(confused_with_cp_strs[0], 16)

        # Skip anything that's not confused with a standard ASCII character.
        # We're only concerned with confusing characters that may be in
        # common identifiers.
        if confused_with_cp >= 128:
            continue

        confused_with_alias = get_alias(confused_with_cp)[0]

        # Skip anything that's not confused with a common or latin character.
        if confused_with_alias not in ('COMMON', 'LATIN'):
            continue

        # Parse and check confusable characters.
        confusable_cp_strs = info['confusable_cp'].split(' ')
        assert len(confusable_cp_strs) == 1, confusable_cp_strs

        confusable_cp = int(confusable_cp_strs[0], 16)

        # There are some confusables, like "1" and "l", that will be in
        # this file. Ignore anything that's ASCII. We only want the more
        # exotic characters.
        if confusable_cp < 128:
            continue

        confusables.append((
            confusable_cp,
            chr(confusable_cp),
            chr(confused_with_cp),
        ))

    confusables.sort()

    return confusables


def build_confusables_file() -> None:
    """Build the confusables file.

    This will generate :file:`reviewboard/codesafety/_unicode_confusables.py`.
    It will contain a dictionary mapping the confusables that we've decided
    to look for in source code.

    Raises:
        RuntimeError:
            This cannot be run on the current version of Python.
    """
    build_categories()
    confusables = update_confusables()

    found_aliases_map: Dict[str, int] = {}

    filename = os.path.abspath(os.path.join(
        __file__, '..', '..', '..', 'reviewboard', 'codesafety',
        '_unicode_confusables.py'))

    with open(filename, 'w') as fp:
        fp.write('# coding: utf-8\n')
        fp.write('#\n')
        fp.write('# THIS FILE IS AUTOMATICALLY GENERATED!\n')
        fp.write('#\n')
        fp.write('# To update this file, run '
                 './contrib/internal/build-confusables.py\n')
        fp.write('\n')
        fp.write('from typing import Dict, Optional, Tuple\n')
        fp.write('\n')
        fp.write('from typing_extensions import TypeAlias\n')
        fp.write('\n')
        fp.write('\n')
        fp.write('ConfusablesMapValue: TypeAlias ='
                 ' Tuple[str, Optional[int]]\n')
        fp.write('ConfusablesMap: TypeAlias ='
                 ' Dict[str, ConfusablesMapValue]\n')
        fp.write('\n')
        fp.write('\n')
        fp.write('COMMON_CONFUSABLES_MAP: ConfusablesMap = {\n')

        alias_index: Optional[int]

        for codepoint, confusable_char, confused_with_char in confusables:
            alias_id, alias_name = get_alias(codepoint)

            if alias_id in DISPLAY_ALIASES:
                assert alias_name

                alias_index = found_aliases_map.setdefault(
                    alias_name, len(found_aliases_map))
            else:
                alias_index = None

            fp.write('    %r: (%r, %r),  # %X; %s\n' %
                     (confusable_char, confused_with_char, alias_index,
                      codepoint, alias_id))

        fp.write('}\n')
        fp.write('\n')
        fp.write('CONFUSABLES_ID_TO_ALIAS_MAP: Tuple[str, ...] = (\n')

        for alias_name in found_aliases_map.keys():
            fp.write('    %r,\n' % alias_name)

        fp.write(')\n')
        fp.write('\n')
        fp.write('CONFUSABLES_ALIAS_TO_ID_MAP: Dict[str, int] = {\n')

        for alias_name, alias_index in found_aliases_map.items():
            fp.write('    %r: %r,\n' % (alias_name, alias_index))

        fp.write('}\n')

    print('Wrote Unicode confusables file: %s' % filename)


if __name__ == '__main__':
    try:
        build_confusables_file()
    except RuntimeError as e:
        sys.stderr.write('%s\n' % e)
        sys.exit(1)



"""Diff viewer view."""

from __future__ import annotations

import itertools
import logging
from typing import Dict, List, Optional, Tuple, TYPE_CHECKING, cast

from django.db.models import Q
from django.utils.translation import gettext
from typing_extensions import NotRequired, TypeAlias, TypedDict

from reviewboard.accounts.mixins import UserProfileRequiredViewMixin
from reviewboard.attachments.models import (FileAttachment,
                                            get_latest_file_attachments)
from reviewboard.diffviewer.commit_utils import get_base_and_tip_commits
from reviewboard.diffviewer.diffutils import DiffFileExtraContext
from reviewboard.diffviewer.models import DiffCommit, DiffSet, FileDiff
from reviewboard.diffviewer.views import (DiffViewerContext,
                                          DiffViewerView,
                                          SerializedDiffContext)
from reviewboard.reviews.context import (ReviewRequestContext,
                                         make_review_request_context,
                                         should_view_draft)
from reviewboard.reviews.ui.diff import DiffReviewUI
from reviewboard.reviews.views.mixins import ReviewRequestViewMixin
from reviewboard.reviews.models import (Review,
                                        ReviewRequestDraft)

if TYPE_CHECKING:
    from datetime import datetime

    from django.http import HttpRequest, HttpResponse

    from reviewboard.reviews.models import Comment, Screenshot
    from reviewboard.reviews.ui.base import SerializedCommentBlocks
    from reviewboard.reviews.ui.diff import SerializedDiffComment

    CommentsDict: TypeAlias = Dict[Tuple[int, Optional[int], Optional[int]],
                                   List[Comment]]


logger = logging.getLogger(__name__)


class ReviewsDiffViewerContext(DiffViewerContext, ReviewRequestContext):
    """Render context for the diff viewer view.

    Version Added:
        7.0
    """

    #: All of the DiffSets connected to the review request.
    diffsets: list[DiffSet]

    #: The current review, if present.
    review: Optional[Review]

    #: The rendered HTML for the review request status.
    review_request_status_html: str

    #: The draft of the review request, if present.
    draft: Optional[ReviewRequestDraft]

    #: The timestamp of the last activity.
    last_activity_time: datetime

    #: The current set of file attachments for the review request.
    file_attachments: list[FileAttachment]

    #: All of the file attachments for the review request.
    #:
    #: This includes all current attachments, as well as old versions and files
    #: that have been removed.
    all_file_attachments: list[FileAttachment]

    #: The screenshots attached to the review request.
    #:
    #: This is a legacy item which has been replaced by the file attachments.
    screenshots: list[Screenshot]

    #: All of the diff comments for the review request.
    comments: CommentsDict


class SerializedDiffsetWithComments(TypedDict):
    """Serialized information about a diff with comments.

    Version Added:
        7.0
    """

    #: Whether this diff is currently being viewed.
    is_current: bool

    #: The revision of the DiffSet.
    revision: int


class SerializedInterdiffsetWithComments(TypedDict):
    """Serialized information about an interdiff with comments.

    Version Added:
        7.0
    """

    #: Whether this interdiff is currently being viewed.
    is_current: bool

    #: The new diff revision in the interdiff pair.
    new_revision: int

    #: The old diff revision in the interdiff pair.
    old_revision: int


class SerializedCommitWithComments(TypedDict):
    """Serialized information about a commit with comments.

    Version Added:
        7.0
    """

    #: The commit ID of the base commit in the commit range.
    base_commit_id: Optional[str]

    #: The PK of the base DiffCommit in the commit range.
    base_commit_pk: Optional[int]

    #: Whether this commit range is currently being viewed.
    is_current: bool

    #: The revision of the diff that contains this commit range.
    revision: int

    #: The commit ID of the tip commit in the commit range.
    tip_commit_id: Optional[str]

    #: The PK of the tip DiffCommit in the commit range.
    tip_commit_pk: Optional[int]


class SerializedCommentsHint(TypedDict):
    """Serialized information about comments in other revisions.

    Version Added:
        7.0
    """

    #: The set of commit ranges that have draft comments.
    commits_with_comments: list[SerializedCommitWithComments]

    #: The set of diffs that have draft comments.
    diffsets_with_comments: list[SerializedDiffsetWithComments]

    #: Whether there are draft comments in other revisions.
    has_other_comments: bool

    #: The set of interdiffs that have draft comments.
    interdiffs_with_comments: list[SerializedInterdiffsetWithComments]


class SerializedReviewsDiffContext(SerializedDiffContext):
    """Serialized diff context information.

    Version Added:
        7.0
    """

    #: The hint to show for draft reviews with comments in other revisions.
    comments_hint: SerializedCommentsHint

    #: The list of files in the current diff view.
    files: list[SerializedReviewsDiffFile]

    #: The number of diff revisions.
    num_diffs: int


class SerializedDiffFileFileDiff(TypedDict):
    """Serialized information about a FileDiff inside the diff files.

    Version Added:
        7.0
    """

    #: The ID of the FileDiff.
    id: int

    #: The revision of the FileDiff.
    revision: int


class SerializedReviewsDiffFile(TypedDict):
    """Serialized information about a file in the diff.

    Version Added:
        7.0
    """

    #: The ID of the base FileDiff when viewing a commit range.
    base_filediff_id: Optional[int]

    #: Whether the file is binary.
    binary: bool

    #: A set of CSS bundle names used to render the file.
    #:
    #: This will only be set for binary files that are rendered through a
    #: :py:class:`~reviewboard.reviews.ui.base.ReviewUI`.
    #:
    #: Version Added:
    #:     8.0
    css_media: NotRequired[set[str]]

    #: Whether the file was deleted in the change.
    deleted: bool

    #: Extra information about the diff file.
    #:
    #: Version Added:
    #:     7.0.4
    extra: DiffFileExtraContext

    #: The ID of the FileDiff.
    id: int

    #: The index of the file in the list of all files.
    index: int

    #: The revision of the interdiff FileDiff, when present.
    interdiff_revision: NotRequired[int]

    #: Information about the interdiff FileDiff, when present.
    interfilediff: NotRequired[SerializedDiffFileFileDiff]

    #: A set of JavaScript bundle names and URLs used to render the file.
    #:
    #: This will only be set for binary files that are rendered through a
    #: :py:class:`~reviewboard.reviews.ui.base.ReviewUI`.
    #:
    #: Version Added:
    #:     8.0
    js_media: NotRequired[set[str]]

    #: Information about the FileDiff.
    filediff: SerializedDiffFileFileDiff

    #: Whether to force rendering an interdiff.
    #:
    #: This is used to ensure that reverted files render correctly.
    force_interdiff: NotRequired[bool]

    #: The filename of the modified version of the file.
    modified_filename: str

    #: The revision of the modified version of the file.
    modified_revision: str

    #: Whether the file is newly-added in the diff.
    newfile: bool

    #: The filename of the original version of the file.
    orig_filename: str

    #: The revision of the original version of the file.
    orig_revision: str

    #: Whether the file is part of a published diff.
    public: bool

    #: The serialized comments already attached to the file's diff.
    serialized_comment_blocks: SerializedCommentBlocks[SerializedDiffComment]


class ReviewsDiffViewerView(ReviewRequestViewMixin,
                            UserProfileRequiredViewMixin,
                            DiffViewerView):
    """Renders the diff viewer for a review request.

    This wraps the base
    :py:class:`~reviewboard.diffviewer.views.DiffViewerView` to display a diff
    for the given review request and the given diff revision or range.

    The view expects the following parameters to be provided:

    ``review_request_id``:
        The ID of the ReviewRequest containing the diff to render.

    The following may also be provided:

    ``revision``:
        The DiffSet revision to render.

    ``interdiff_revision``:
        The second DiffSet revision in an interdiff revision range.

    ``local_site``:
        The LocalSite the ReviewRequest must be on, if any.

    See :py:class:`~reviewboard.diffviewer.views.DiffViewerView`'s
    documentation for the accepted query parameters.
    """

    def __init__(
        self,
        **kwargs,
    ) -> None:
        """Initialize a view for the request.

        Args:
            **kwargs (dict):
                Keyword arguments passed to :py:meth:`as_view`.
        """
        super().__init__(**kwargs)

        self.draft = None
        self.interdiffset = None

    def get(
        self,
        request: HttpRequest,
        revision: Optional[int] = None,
        interdiff_revision: Optional[int] = None,
        *args,
        **kwargs,
    ) -> HttpResponse:
        """Handle HTTP GET requests for this view.

        This will look up the review request and DiffSets, given the
        provided information, and pass them to the parent class for rendering.

        Args:
            request (django.http.HttpRequest):
                The HTTP request from the client.

            revision (int, optional):
                The revision of the diff to view. This defaults to the latest
                diff.

            interdiff_revision (int, optional):
                The revision to use for an interdiff, if viewing an interdiff.

            *args (tuple):
                Positional arguments passed to the handler.

            **kwargs (dict):
                Keyword arguments passed to the handler.

        Returns:
            django.http.HttpResponse:
            The HTTP response to send to the client.
        """
        review_request = self.review_request

        self.draft = review_request.get_draft(user=self.request.user)

        self.diffset = self.get_diff(revision, self.draft)

        if interdiff_revision and interdiff_revision != revision:
            # An interdiff revision was specified. Try to find a matching
            # diffset.
            self.interdiffset = self.get_diff(interdiff_revision, self.draft)

        return super().get(
            request=request,
            diffset=self.diffset,
            interdiffset=self.interdiffset,
            *args,
            **kwargs)

    def get_context_data(
        self,
        diffset: DiffSet,
        interdiffset: Optional[DiffSet],
        **kwargs,
    ) -> ReviewsDiffViewerContext:
        """Return additional context data for the template.

        This provides some additional data used for rendering the diff
        viewer. This data is more specific to the reviewing functionality,
        as opposed to the data calculated by
        :py:meth:`DiffViewerView.get_context_data
        <reviewboard.diffviewer.views.DiffViewerView.get_context_data>`
        which is more focused on the actual diff.

        Args:
            diffset (reviewboard.diffviewer.models.DiffSet):
                The diffset being viewed.

            interdiffset (reviewboard.diffviewer.models.DiffSet):
                The interdiff diffset, if present.

            **kwargs (dict):
                Keyword arguments passed to the handler.

        Returns:
            dict:
            Context data used to render the template.
        """
        request = self.request
        review_request = self.review_request
        draft = self.draft
        draft_diffset: Optional[DiffSet] = None

        # Only include the draft diffset if the user is the owner of the
        # review request, or if the user wants to view the unpublished draft.
        if should_view_draft(request=request, review_request=review_request,
                             draft=draft):
            if draft.diffset:
                draft_diffset = draft and draft.diffset

            review_request_details = draft or review_request
        else:
            review_request_details = review_request

        # Try to find an existing pending review of this diff from the
        # current user.
        pending_review = review_request.get_pending_review(request.user)

        is_draft_diff = draft_diffset == self.diffset
        is_draft_interdiff = bool(self.interdiffset and
                                  draft_diffset == self.interdiffset)

        # Get the list of diffsets. We only want to calculate this once.
        diffsets = review_request.get_diffsets()

        # Get this before we add the draft diffset to the list. Otherwise the
        # timestamp won't be correct.
        last_activity_time = review_request.get_last_activity_info(
            diffsets)['timestamp']

        if draft_diffset:
            diffsets.append(draft_diffset)

        num_diffs = len(diffsets)

        if num_diffs > 0:
            latest_diffset = diffsets[-1]
        else:
            latest_diffset = None

        # We'll need this for later lookups. The diffsets returned by
        # get_diffsets() already have the filediffs pre-fetched.
        filediffs_by_id: dict[int, FileDiff] = {}

        for ds in diffsets:
            for filediff in ds.files.all():
                filediffs_by_id[filediff.pk] = filediff

        file_attachments = list(review_request_details.get_file_attachments())
        screenshots = list(review_request_details.get_screenshots())

        latest_file_attachments = get_latest_file_attachments(file_attachments)
        social_page_image_url = self.get_social_page_image_url(
            latest_file_attachments)

        # Since we have these attachments, we can use them to avoid database
        # lookups when fetching review UIs for binary file diffs below.
        attachments_by_filediff_id: dict[int, FileAttachment] = {}

        for attachment in file_attachments:
            if (f_id := attachment.added_in_filediff_id):
                attachments_by_filediff_id[f_id] = attachment

        # Build the status information shown below the summary.
        close_info = review_request.get_close_info()

        if latest_diffset:
            status_extra_info = [{
                'text': gettext('Latest diff uploaded {timestamp}'),
                'timestamp': latest_diffset.timestamp,
            }]
        else:
            status_extra_info = []

        review_request_status_html = self.get_review_request_status_html(
            review_request_details=review_request_details,
            close_info=close_info,
            extra_info=status_extra_info)

        # Compute the lists of comments based on filediffs and interfilediffs.
        # We do this using the 'through' table so that we can select_related
        # the reviews and comments.
        comments: CommentsDict = {}
        q = (
            Review.comments.through.objects
            .filter(review__review_request=review_request)
            .select_related()
        )

        for obj in q:
            comment = obj.comment
            comment.review_obj = obj.review
            key = (comment.filediff_id, comment.interfilediff_id,
                   comment.base_filediff_id)
            comments.setdefault(key, []).append(comment)

        base_commit_id: Optional[int] = None
        base_commit: Optional[DiffCommit] = None
        tip_commit_id: Optional[int] = None
        tip_commit: Optional[DiffCommit] = None

        if diffset.commit_count and not interdiffset:
            # Base and tip commit selection is not supported in interdiffs.
            raw_base_commit_id = request.GET.get('base-commit-id')
            raw_tip_commit_id = request.GET.get('tip-commit-id')

            if raw_base_commit_id is not None:
                try:
                    base_commit_id = int(raw_base_commit_id)
                except ValueError:
                    pass

            if raw_tip_commit_id is not None:
                try:
                    tip_commit_id = int(raw_tip_commit_id)
                except ValueError:
                    pass

        all_commits: list[DiffCommit] = list(DiffCommit.objects.filter(
            Q(diffset__history=review_request.diffset_history_id) |
            Q(diffset__review_request_draft__review_request=review_request)))

        if base_commit_id or tip_commit_id:
            base_commit, tip_commit = get_base_and_tip_commits(
                base_commit_id,
                tip_commit_id,
                commits=all_commits)

        # Build the final context for the page.
        context = cast(
            ReviewsDiffViewerContext,
            super().get_context_data(
                diffset=diffset,
                interdiffset=interdiffset,
                all_commits=all_commits,
                base_commit=base_commit,
                tip_commit=tip_commit,
                **kwargs))

        context.update({
            'all_file_attachments': file_attachments,
            'comments': comments,
            'diffsets': diffsets,
            'file_attachments': latest_file_attachments,
            'last_activity_time': last_activity_time,
            'review': pending_review,
            'review_request_status_html': review_request_status_html,
            'screenshots': screenshots,
        })

        context.update(make_review_request_context(
            request=request,
            review_request=review_request,
            close_info=close_info,
            draft=draft,
            review_request_details=review_request_details,
            is_diff_view=True,
            social_page_title=(
                f'Diff for Review Request #{review_request.display_id}: '
                f'{review_request_details.summary}'),
            social_page_image_url=social_page_image_url))

        diff_context = cast(SerializedReviewsDiffContext,
                            context['diff_context'])

        diff_context.update({
            'num_diffs': num_diffs,
            'comments_hint': self._get_comments_hint(
                pending_review=pending_review,
                all_comments=comments,
                all_diffsets=diffsets,
                all_commits=all_commits,
                filediffs_by_id=filediffs_by_id,
                current_diffset=diffset,
                current_interdiffset=interdiffset,
                current_base_commit_id=base_commit_id,
                current_tip_commit_id=tip_commit_id,
            ),
        })
        diff_context['revision'].update({
            'latest_revision': (latest_diffset.revision
                                if latest_diffset else None),
            'is_draft_diff': is_draft_diff,
            'is_draft_interdiff': is_draft_interdiff,
        })

        files: list[SerializedReviewsDiffFile] = []

        for f in context['files']:
            filediff = f['filediff']
            interfilediff = f['interfilediff']
            base_filediff = f['base_filediff']
            binary = f['binary']
            filediff_id = filediff.pk

            interfilediff_id: Optional[int] = None
            base_filediff_id: Optional[int] = None

            if base_filediff:
                base_filediff_id = base_filediff.pk

            if interfilediff:
                interfilediff_id = interfilediff.pk

            key = (filediff_id, interfilediff_id, base_filediff_id)

            file_comments = comments.get(key, [])

            review_ui = DiffReviewUI(
                review_request=review_request,
                obj=filediff,
                base_filediff=base_filediff,
                interfilediff=interfilediff,
                request=self.request)

            data: SerializedReviewsDiffFile = {
                'base_filediff_id': base_filediff_id,
                'binary': binary,
                'deleted': f['deleted'],
                'extra': f['extra'],
                'id': filediff_id,
                'index': f['index'],
                'filediff': {
                    'id': filediff_id,
                    'revision': filediff.diffset.revision,
                },
                'modified_filename': f['modified_filename'],
                'modified_revision': f['modified_revision'],
                'newfile': f['newfile'],
                'orig_filename': f['orig_filename'],
                'orig_revision': f['orig_revision'],
                'public': f['public'],
                'serialized_comment_blocks':
                    review_ui.serialize_comments(file_comments),
            }

            if interfilediff:
                data['interfilediff'] = {
                    'id': interfilediff.pk,
                    'revision': interfilediff.diffset.revision,
                }

            if f['force_interdiff']:
                data['force_interdiff'] = True
                data['interdiff_revision'] = f['force_interdiff_revision']

            if binary:
                # Get the static media bundle names used in rendering the diff
                # file. We use this to make sure that static media only gets
                # added to the page once per review UI type in the diff viewer.
                try:
                    attachment = attachments_by_filediff_id[filediff_id]
                except KeyError:
                    # This file diff is from the cumulative diff. Need to hit
                    # the database instead.
                    attachment = FileAttachment.objects.get_for_filediff(
                        filediff)

                if attachment and (review_ui := attachment.review_ui):
                    data['js_media'] = set(review_ui.js_bundle_names +
                                           review_ui.js_files)
                    data['css_media'] = set(review_ui.css_bundle_names)

            files.append(data)

        diff_context['files'] = files

        return context

    def _get_comments_hint(
        self,
        *,
        pending_review: Optional[Review],
        all_comments: CommentsDict,
        all_diffsets: list[DiffSet],
        all_commits: list[DiffCommit],
        filediffs_by_id: dict[int, FileDiff],
        current_diffset: DiffSet,
        current_interdiffset: Optional[DiffSet],
        current_base_commit_id: Optional[int],
        current_tip_commit_id: Optional[int],
    ) -> SerializedCommentsHint:
        """Return the comments hint for the diff viewer.

        Args:
            pending_review (reviewboard.reviews.models.Review):
                The user's pending review, if any.

            all_comments (dict):
                All diff comments on the review request, sorted by which
                revision they're on.

            all_diffsets (list of reviewboard.diffviewer.models.DiffSet):
                All DiffSets on the review request.

            filediffs_by_id (dict):
                A mapping from FileDiff PK to the object.

            current_diffset (reviewboard.diffviewer.models.DiffSet):
                The current diffset being viewed.

            current_interdiffset (reviewboard.diffviewer.models.DiffSet):
                The current interdiffset, if the user is viewing an interdiff.

            current_base_commit (reviewboard.diffviewer.models.DiffCommit):
                The current base commit, if the user is viewing a commit range.

            current_tip_commit (reviewboard.diffviewer.models.DiffCommit):
                The current tip commit, if the user is viewing a commit range.

        Returns:
            SerializedCommentsHint:
            The comments hint to send to the client UI.
        """
        hint: SerializedCommentsHint = {
            'commits_with_comments': [],
            'diffsets_with_comments': [],
            'has_other_comments': False,
            'interdiffs_with_comments': [],
        }

        if pending_review is None:
            return hint

        pending_comments: list[Comment] = []

        for comment in itertools.chain(*all_comments.values()):
            review = comment.get_review()

            if review.pk == pending_review.pk:
                pending_comments.append(comment)

        diffsets_by_id: dict[int, DiffSet] = {
            diffset.pk: diffset
            for diffset in all_diffsets
        }

        # The diffsets already have the filediffs pre-fetched.
        filediffs_by_id: dict[int, FileDiff] = {}

        for diffset in all_diffsets:
            for filediff in diffset.files.all():
                filediffs_by_id[filediff.pk] = filediff

        # A set of DiffSet IDs for diffs that have comments.
        diffset_ids_with_comments: set[int] = set()

        # A set of ID pairs for interdiffs that have comments.
        interdiffset_ids_with_comments: set[tuple[int, int]] = set()

        # A set of DiffSet ID plus two DiffCommit IDs for comments on commit
        # ranges.
        commit_ranges_with_comments: set[tuple[int, int, int]] = \
            set()

        # Value to use when the base commit ID for a comment is None, so that
        # commit_ranges_with_comments is sortable.
        NO_BASE_COMMIT = -1

        # Go through all diff comments and sort them into our *_with_comments
        # sets.
        for comment in pending_comments:
            filediff = filediffs_by_id[comment.filediff_id]
            diffset_id = filediff.diffset_id

            if comment.interfilediff_id is not None:
                interfilediff = filediffs_by_id[comment.interfilediff_id]
                interdiffset_ids_with_comments.add(
                    (diffset_id, interfilediff.diffset_id))
            elif filediff.commit_id is None:
                diffset_ids_with_comments.add(diffset_id)
            else:
                tip_commit_id = filediff.commit_id
                base_filediff_id = comment.base_filediff_id

                if base_filediff_id is None:
                    base_commit_id = NO_BASE_COMMIT
                else:
                    base_filediff = filediffs_by_id[base_filediff_id]
                    base_commit_id = base_filediff.commit_id

                commit_ranges_with_comments.add(
                    (diffset_id, base_commit_id, tip_commit_id))

        # Now go through file attachment comments and add any that correspond
        # to diffs.
        for comment in pending_review.file_attachment_comments.all():
            revision_info = comment.get_comment_diff_revision_info()

            if not revision_info:
                continue

            diffset_id = revision_info['diffset_id']
            interdiffset_id = revision_info['interdiffset_id']
            base_commit_id = revision_info['base_commit_id']
            tip_commit_id = revision_info['tip_commit_id']

            if interdiffset_id is not None:
                interdiffset_ids_with_comments.add(
                    (diffset_id, interdiffset_id)
                )
            elif base_commit_id is not None or tip_commit_id is not None:
                if base_commit_id is None:
                    base_commit_id = NO_BASE_COMMIT

                if tip_commit_id is None:
                    tip_commit_id = revision_info.get('last_commit_id')
                    assert tip_commit_id is not None

                commit_ranges_with_comments.add(
                    (diffset_id, base_commit_id, tip_commit_id)
                )
            else:
                diffset_ids_with_comments.add(revision_info['diffset_id'])

        commits_by_id: dict[int, DiffCommit] = {
            commit.pk: commit
            for commit in all_commits
        }

        has_other_comments: bool = False

        # Find all diffsets that have comments.
        current_is_diffset = (
            current_interdiffset is None and
            current_base_commit_id is None and
            current_tip_commit_id is None)
        diffsets_with_comments: list[SerializedDiffsetWithComments] = []

        for diffset_id in sorted(diffset_ids_with_comments):
            is_current = (current_is_diffset and
                          current_diffset.pk == diffset_id)

            if not is_current:
                has_other_comments = True

            diffsets_with_comments.append({
                'is_current': is_current,
                'revision': diffsets_by_id[diffset_id].revision,
            })

        # Now find all interdiffs that have comments.
        current_is_interdiffset = current_interdiffset is not None
        interdiffsets_with_comments: \
            list[SerializedInterdiffsetWithComments] = []

        for base_id, tip_id in sorted(interdiffset_ids_with_comments):
            base_diffset = diffsets_by_id[base_id]
            tip_diffset = diffsets_by_id[tip_id]

            is_current = (current_is_interdiffset and
                          current_diffset.pk == base_id and
                          current_interdiffset.pk == tip_id)

            if not is_current:
                has_other_comments = True

            interdiffsets_with_comments.append({
                'is_current': is_current,
                'new_revision': tip_diffset.revision,
                'old_revision': base_diffset.revision,
            })

        # Now find all commit ranges that have comments.
        current_is_commit_range = (current_base_commit_id is not None or
                                   current_tip_commit_id is not None)
        commits_with_comments: list[SerializedCommitWithComments] = []

        for diffset_id, base_commit_id, tip_commit_id in \
            sorted(commit_ranges_with_comments):
            diffset = diffsets_by_id[diffset_id]

            if base_commit_id == NO_BASE_COMMIT:
                # An unspecified base commit ID in the comment means that the
                # selected range started at the first commit in the series for
                # the diffset.
                base_commit: DiffCommit

                for commit in all_commits:
                    if commit.parent_id == diffset.base_commit_id:
                        base_commit = commit
                        break
                else:
                    logger.error('Unable to find base commit for diffset %s',
                                 diffset_id)
                    continue

                base_commit_pk = None
            else:
                base_commit = commits_by_id[base_commit_id]
                base_commit_pk = base_commit.pk

            tip_commit = commits_by_id[tip_commit_id]

            is_current = (
                current_is_commit_range and
                current_diffset.pk == diffset_id and
                ((current_base_commit_id is None and
                  base_commit_id == NO_BASE_COMMIT) or
                 (current_base_commit_id is not None and
                  current_base_commit_id == base_commit_id)) and
                current_tip_commit_id == tip_commit_id)

            if not is_current:
                has_other_comments = True

            # Comments on commits are a little bit odd in that the
            # base_filediff_id and corresponding base_commit_id are actually
            # the parent commit of the one which is selected in the range
            # selector. We therefore want to serialize the PK of the
            # base_commit but the commit_id of the child of the base_commit
            base_commit_commit_id: Optional[str] = None

            if base_commit_id is NO_BASE_COMMIT:
                base_commit_commit_id = base_commit.commit_id
            else:
                for commit in all_commits:
                    if commit.parent_id == base_commit.commit_id:
                        base_commit_commit_id = commit.commit_id
                        break
                else:
                    logger.error('Unable to find child commit for comment '
                                 'base commit pk=%s',
                                 base_commit_id)

            commits_with_comments.append({
                'base_commit_id': base_commit_commit_id,
                'base_commit_pk': base_commit_pk,
                'revision': diffset.revision,
                'is_current': is_current,
                'tip_commit_id': tip_commit.commit_id,
                'tip_commit_pk': tip_commit.pk,
            })

        hint['diffsets_with_comments'] = diffsets_with_comments
        hint['commits_with_comments'] = commits_with_comments
        hint['has_other_comments'] = has_other_comments
        hint['interdiffs_with_comments'] = interdiffsets_with_comments

        return hint



"""Unit tests for reviewboard.accounts.views.edit_oauth_app."""

from django.contrib.auth.models import User
from django.urls import reverse

from reviewboard.testing import TestCase


class EditOAuthAppViewTests(TestCase):
    """Unit tests for edit_oauth_app."""

    fixtures = ['test_users']

    def test_get(self):
        """Testing edit_oauth_app GET"""
        self.client.login(username='doc', password='doc')

        user = User.objects.get(username='doc')
        app = self.create_oauth_application(user=user)
        url = reverse('edit-oauth-app', kwargs={
            'app_id': app.pk,
        })

        response = self.client.get(url)

        self.assertEqual(response.status_code, 200)
        self.assertTemplateUsed(response, 'accounts/edit_oauth_app.html')
        self.assertEqual(response.context['app'], app)



"""Columns for datagrids."""

from __future__ import annotations

from typing import TYPE_CHECKING

from django.db.models import Count, Q
from django.template.defaultfilters import date
from django.urls import NoReverseMatch
from django.utils.html import escape, format_html
from django.utils.safestring import mark_safe
from django.utils.translation import gettext, gettext_lazy as _, ngettext
from djblets.datagrid.grids import CheckboxColumn, Column, DateTimeColumn
from djblets.siteconfig.models import SiteConfiguration

from reviewboard.accounts.models import ReviewRequestVisit
from reviewboard.avatars import avatar_services
from reviewboard.reviews.models import Group, ReviewRequest
from reviewboard.reviews.templatetags.reviewtags import render_star
from reviewboard.site.urlresolvers import local_site_reverse

if TYPE_CHECKING:
    from collections.abc import Mapping, Sequence
    from datetime import datetime
    from typing import Any, Final, NotRequired, TypedDict

    from django.contrib.auth.models import User
    from django.db.models import QuerySet
    from django.utils.safestring import SafeString
    from djblets.datagrid.grids import StatefulColumn, _RenderContext
    from typelets.django.strings import StrPromise

    from reviewboard.accounts.models import (StarrableObject,
                                             User as RBUser)
    from reviewboard.reviews.models import Review

    class _BugsColumnData(TypedDict):
        id: str
        url: NotRequired[str]


class BaseStarColumn(Column):
    """Indicates if an item is starred.

    This is the base class for all columns that deal with starring items.

    The star is interactive, allowing the user to star or unstar the item.
    """

    detailed_label = _('Starred')
    image_alt = _('Starred')
    image_class = 'rb-icon rb-icon-star-on'
    shrink = True

    #: The starrable model handled by the column.
    starrable_model: type[StarrableObject]

    def augment_queryset(
        self,
        state: StatefulColumn,
        queryset: QuerySet,
    ) -> QuerySet:
        """Add additional queries to the queryset.

        This will load the list of starrable object IDs to check in the
        starred list, and then set them in the user's star cache for
        later rendering.

        Args:
            state (djblets.datagrid.grids.StatefulColumn):
                The column state.

            queryset (django.db.models.query.QuerySet):
                The queryset to augment.

        Returns:
            django.db.models.query.QuerySet:
            The resulting queryset.
        """
        user = state.datagrid.request.user

        if user.is_authenticated:
            profile = user.get_profile()
            profile.prefetch_starred_objects(self.starrable_model,
                                             state.datagrid.id_list)

        return queryset

    def render_data(
        self,
        state: StatefulColumn,
        obj: StarrableObject,
    ) -> SafeString:
        """Return the rendered contents of the column.

        Args:
            state (djblets.datagrids.grids.StatefulColumn):
                The state for the DataGrid instance.

            obj (django.db.models.Model):
                The starrable model object for the row.

        Returns:
            django.utils.safestring.SafeString:
            The rendered data as HTML.
        """
        return render_star(state.datagrid.request.user, obj)

    def get_raw_object_value(
        self,
        state: StatefulColumn,
        obj: StarrableObject,
    ) -> bool:
        """Return the raw value for the star state.

        This will determine whether the object is starred and save it
        for lookup during rendering.

        Version Added:
            8.0

        Args:
            state (djblets.datagrids.grids.StatefulColumn):
                The state for the DataGrid instance.

            obj (django.db.models.Model):
                The starrable model object for the row.

        Returns:
            bool:
            The star state for the object.
        """
        user = state.datagrid.request.user

        return (
            user.is_authenticated and
            user.get_profile().is_object_starred(obj)
        )


class UsernameColumn(Column):
    """A column for showing a username and the user's avatar.

    The username and avatar will link to the user's profile page and will
    show basic profile information when hovering over the link.

    When constructing an instance of this column, the relation between the
    object being represented in the datagrid and the user can be specified
    as a tuple or list of field names forming a path to the user field.
    """

    AVATAR_SIZE = 24

    label = _('Username')
    css_class = 'submitter-column'
    shrink = True
    sortable = True
    link = True
    link_css_class = 'user'

    def __init__(
        self,
        user_relation: (Sequence[str] | None) = None,
        *args,
        **kwargs,
    ) -> None:
        """Initialize the column.

        Args:
            user_relation (list of str, optional):
                A list of fields forming a relation path to the user. This can
                be left blank if representing the user.

            *args (tuple):
                Additional positional arguments to pass to the column.

            **kwargs (dict):
                Additional keyword arguments to pass to the column.
        """
        if user_relation is None:
            user_relation = []

        self._user_relation = user_relation

        super().__init__(
            db_field='__'.join([*user_relation, 'username']),
            link_func=self._link_user,
            *args, **kwargs)

    def get_user(
        self,
        obj: Any,
    ) -> RBUser:
        """Return the user associated with this object.

        Args:
            obj (object):
                The object provided to the column.

        Returns:
            reviewboard.accounts.models.User:
            The resulting user.
        """
        # Look up the user in the provided obj by traversing the relation.
        # If _user_relation is empty, then obj is the user.
        user = obj

        for field_name in self._user_relation:
            user = getattr(user, field_name)

        return user

    def render_data(
        self,
        state: StatefulColumn,
        obj: Any,
    ) -> SafeString:
        """Render the user's name and avatar as HTML.

        Args:
            state (djblets.datagrid.grids.StatefulColumn):
                The column state.

            obj (django.db.models.Model):
                The object being rendered in the datagrid.

        Returns:
            django.utils.safestring.SafeString:
            The HTML for the column.
        """
        user = self.get_user(obj)

        # If avatars are eanbled, we'll want to include that in the resulting
        # HTML.
        siteconfig = SiteConfiguration.objects.get_current()
        request = state.datagrid.request
        avatar_html = ''

        if siteconfig.get(avatar_services.AVATARS_ENABLED_KEY):
            avatar_service = avatar_services.for_user(user)

            if avatar_service:
                avatar_html = avatar_service.render(request=request,
                                                    user=user,
                                                    size=self.AVATAR_SIZE)

        # Render the link to the user page, using the avatar and username.
        username = user.username

        return format_html('{0}{1}', avatar_html, username)

    def augment_queryset(
        self,
        state: StatefulColumn,
        queryset: QuerySet,
    ) -> QuerySet:
        """Add additional queries to the queryset.

        This will select fields for the user and the user's profile, to
        help with query performance.

        Args:
            state (djblets.datagrid.grids.StatefulColumn):
                The column state.

            queryset (django.db.models.query.QuerySet):
                The queryset to augment.

        Returns:
            django.db.models.query.QuerySet:
            The resulting queryset.
        """
        user_field = '__'.join(self._user_relation)

        if user_field:
            fields = [user_field, '%s__profile' % user_field]
        else:
            fields = ['profile']

        return queryset.select_related(*fields)

    def _link_user(
        self,
        state: StatefulColumn,
        obj: Any,
        *args,
    ) -> str:
        """Return the URL to link the user associated with this object.

        Args:
            state (djblets.datagrid.grids.StatefulColumn, unused):
                The column state.

            obj (object):
                The object provided to the column.

            *args (tuple):
                Additional keyword arguments provided to the method.

        Returns:
            str:
            The URL for the user.
        """
        return local_site_reverse(
            'user',
            request=state.datagrid.request,
            kwargs={
                'username': self.get_user(obj).username,
            })


class FullNameColumn(Column):
    """Shows the full name of the user when appropriate."""

    def augment_queryset(
        self,
        state: StatefulColumn,
        queryset: QuerySet,
    ) -> QuerySet:
        """Add additional queries to the queryset.

        This will select fields for the user and the user's profile, to
        help with query performance.

        Args:
            state (djblets.datagrid.grids.StatefulColumn):
                The column state.

            queryset (django.db.models.query.QuerySet):
                The queryset to augment.

        Returns:
            django.db.models.query.QuerySet:
            The resulting queryset.
        """
        return queryset.select_related('profile')

    def get_raw_object_value(
        self,
        state: StatefulColumn,
        obj: RBUser,
    ) -> str | None:
        """Return the raw value for the full name.

        If the profile is private, the full name will not be provided.

        Version Added:
            8.0

        Args:
            state (djblets.datagrids.grids.StatefulColumn):
                The state for the DataGrid instance.

            obj (reviewboard.accounts.models.User):
                The user for the row.

        Returns:
            str:
            The full name, or ``None`` if the profile is private.
        """
        profile = obj.get_profile()
        request_user = state.datagrid.request.user

        if obj.is_profile_visible(request_user):
            return profile.get_display_name(request_user)

        return None


class BugsColumn(Column):
    """Shows the list of bugs specified on a review request.

    The list of bugs will be linked to the bug tracker, if a bug tracker
    was configured for the repository the review request's change is on.
    """

    label = _('Bugs')
    css_class = 'bugs'
    shrink = True
    sortable = False
    link = False

    def augment_queryset(
        self,
        state: StatefulColumn,
        queryset: QuerySet,
    ) -> QuerySet:
        """Add additional queries to the queryset.

        This will include the repository along with each review request.

        Args:
            state (djblets.datagrid.grids.StatefulColumn):
                The column state.

            queryset (django.db.models.query.QuerySet):
                The queryset to augment.

        Returns:
            django.db.models.query.QuerySet:
            The resulting queryset.
        """
        return queryset.select_related('repository')

    def render_data(
        self,
        state: StatefulColumn,
        obj: ReviewRequest,
    ) -> SafeString:
        """Return the rendered contents of the column.

        Args:
            state (djblets.datagrid.grids.StatefulColumn):
                The state for the DataGrid instance.

            obj (reviewboard.reviews.models.ReviewRequest):
                The review request being rendered for this row.

        Returns:
            django.utils.safestring.SafeString:
            The rendered HTML.
        """
        result: list[SafeString] = []

        for bug_info in self.get_raw_object_value(state, obj):
            url = bug_info.get('url')
            bug_id = bug_info['id']

            if url:
                html = format_html(
                    '<a class="bug" href="{}">{}</a>',
                    url, bug_id,
                )
            else:
                html = format_html(
                    '<span class="bug">{}</span>',
                    bug_id,
                )

            result.append(html)

        return mark_safe(', '.join(result))

    def get_raw_object_value(
        self,
        state: StatefulColumn,
        obj: ReviewRequest,
    ) -> Sequence[_BugsColumnData]:
        """Return the raw value for the bugs information.

        This will return the list of bug IDs and URLs (if present) from the
        review request.

        Version Added:
            8.0

        Args:
            state (djblets.datagrids.grids.StatefulColumn):
                The state for the DataGrid instance.

            obj (reviewboard.reviews.models.ReviewRequest):
                The review request for the row.

        Returns:
            list of _BugsColumnData:
            The list of bug information.
        """
        review_request = obj
        repository = review_request.repository

        # Determine the Local Site name for the review request, if any.
        local_site_name: (str | None) = None

        if review_request.local_site:
            local_site_name = review_request.local_site.name

        result: list[_BugsColumnData]
        bugs = obj.get_bug_list()

        if repository and repository.bug_tracker:
            result = []

            for bug in bugs:
                bug_info: _BugsColumnData = {
                    'id': bug,
                }

                try:
                    bug_info['url'] = local_site_reverse(
                        'bug_url',
                        local_site_name=local_site_name,
                        args=[review_request.display_id, bug],
                    )
                except NoReverseMatch:
                    pass

                result.append(bug_info)
        else:
            result = [
                {
                    'id': bug,
                }
                for bug in bugs
            ]

        return result


class ReviewRequestCheckboxColumn(CheckboxColumn):
    """A column containing a check-box."""

    def render_data(
        self,
        state: StatefulColumn,
        obj: ReviewRequest,
    ) -> SafeString:
        """Return the rendered contents of the column.

        Args:
            state (djblets.datagrid.grids.StatefulColumn):
                The state for the DataGrid instance.

            obj (reviewboard.reviews.models.ReviewRequest):
                The review request being rendered for this row.

        Returns:
            django.utils.safestring.SafeString:
            The rendered HTML.
        """
        if not self.is_selectable(state, obj):
            return mark_safe('')

        if self.is_selected(state, obj):
            checked = mark_safe('checked="true"')
        else:
            checked = mark_safe('')

        return format_html(
            '<input type="checkbox" data-object-id="{}" '
            'data-checkbox-name="{}" {}>',
            obj.display_id,
            self.checkbox_name,
            checked,
        )


class DateTimeSinceColumn(DateTimeColumn):
    """Displays how long it has been since a given date/time.

    These columns will dynamically update as the page is shown, so that the
    number of minutes, hours, days, etc. ago is correct.
    """

    def render_data(
        self,
        state: StatefulColumn,
        obj: Any,
    ) -> SafeString:
        """Return the rendered contents of the column.

        Args:
            state (djblets.datagrid.grids.StatefulColumn):
                The state for the DataGrid instance.

            obj (object):
                The object being rendered for this row.

        Returns:
            django.utils.safestring.SafeString:
            The rendered HTML.
        """
        value = self.get_raw_object_value(state, obj)

        if value is None:
            return mark_safe('')

        return format_html(
            '<time class="timesince" datetime="{}">{}</time>',
            date(value, 'c'),
            super().render_data(state, obj),
        )


class DiffUpdatedColumn(DateTimeColumn):
    """Shows the date/time that the diff was last updated."""

    label = _('Diff Updated')
    db_field = 'diffset_history__last_diff_updated'
    sortable = True
    link = False

    def augment_queryset(
        self,
        state: StatefulColumn,
        queryset: QuerySet,
    ) -> QuerySet:
        """Add additional queries to the queryset.

        This will include the diffset history along with each review request.

        Args:
            state (djblets.datagrid.grids.StatefulColumn):
                The column state.

            queryset (django.db.models.query.QuerySet):
                The queryset to augment.

        Returns:
            django.db.models.query.QuerySet:
            The resulting queryset.
        """
        return queryset.select_related('diffset_history')

    def get_raw_object_value(
        self,
        state: StatefulColumn,
        obj: ReviewRequest,
    ) -> datetime | None:
        """Return the diff updated timestamp.

        Version Added:
            8.0

        Args:
            state (djblets.datagrids.grids.StatefulColumn):
                The state for the DataGrid instance.

            obj (reviewboard.reviews.models.ReviewRequest):
                The review request to return the value for.

        Returns:
            datetime.datetime:
            The diff updated timestamp, or ``None`` if there's no diff.
        """
        if ((diffset_history := obj.diffset_history) and
            (last_diff_updated := diffset_history.last_diff_updated)):
            return last_diff_updated

        return None


class DiffUpdatedSinceColumn(DateTimeSinceColumn):
    """Shows the elapsed time since the diff was last updated."""

    label = _('Diff Updated')
    db_field = 'diffset_history__last_diff_updated'
    sortable = True
    link = False

    def augment_queryset(
        self,
        state: StatefulColumn,
        queryset: QuerySet,
    ) -> QuerySet:
        """Add additional queries to the queryset.

        This will include the diffset history along with each review request.

        Args:
            state (djblets.datagrid.grids.StatefulColumn):
                The column state.

            queryset (django.db.models.query.QuerySet):
                The queryset to augment.

        Returns:
            django.db.models.query.QuerySet:
            The resulting queryset.
        """
        return queryset.select_related('diffset_history')

    def get_raw_object_value(
        self,
        state: StatefulColumn,
        obj: ReviewRequest,
    ) -> datetime | None:
        """Return the diff updated timestamp.

        Version Added:
            8.0

        Args:
            state (djblets.datagrids.grids.StatefulColumn):
                The state for the DataGrid instance.

            obj (reviewboard.reviews.models.ReviewRequest):
                The review request to return the value for.

        Returns:
            datetime.datetime:
            The diff updated timestamp, or ``None`` if there's no diff.
        """
        if obj.diffset_history and obj.diffset_history.last_diff_updated:
            return obj.diffset_history.last_diff_updated

        return None


class GroupMemberCountColumn(Column):
    """Shows the number of users that are part of a review group."""

    link = True

    def augment_queryset_for_data(
        self,
        state: StatefulColumn,
        queryset: QuerySet,
        **kwargs,
    ) -> QuerySet:
        """Augment a queryset for data-rendering purposes.

        This will count the number of group members for display.

        Version Added:
            5.0.7

        Args:
            state (djblets.datagrid.grids.StatefulColumn):
                The state for the DataGrid instance.

            queryset (django.db.models.query.QuerySet):
                The queryset to augment.

            **kwargs (dict):
                Additional keyword arguments for future expansion.

        Returns:
            django.db.models.query.QuerySet:
            The resulting augmented QuerySet.
        """
        return queryset.annotate(column_group_member_count=Count('users'))

    def render_data(
        self,
        state: StatefulColumn,
        obj: Group,
    ) -> SafeString:
        """Return the rendered contents of the column.

        Args:
            state (djblets.datagrid.grids.StatefulColumn):
                The state for the DataGrid instance.

            obj (reviewboard.reviews.models.group.Group):
                The object being rendered for this row.

        Returns:
            django.utils.safestring.SafeString:
            The rendered HTML.
        """
        return escape(str(self.get_raw_object_value(state, obj)))

    def get_raw_object_value(
        self,
        state: StatefulColumn,
        obj: Group,
    ) -> int:
        """Return the raw value for the group membership count.

        Version Added:
            8.0

        Args:
            state (djblets.datagrids.grids.StatefulColumn):
                The state for the DataGrid instance.

            obj (reviewboard.reviews.models.Group):
                The review group for the row.

        Returns:
            int:
            The group membership count.
        """
        return getattr(obj, 'column_group_member_count', 0)

    @staticmethod
    def link_func(
        state: StatefulColumn,
        obj: Group,
        value: str,
    ) -> str:
        """Return the link to the object in the column.

        Args:
            state (djblets.datagrid.grids.StatefulColumn):
                The column state.

            obj (reviewboard.reviews.models.Group):
                The review group for the row.

            value (str):
                The rendered data.

        Returns:
            str:
            The URL to link to.
        """
        return local_site_reverse('group-members',
                                  request=state.datagrid.request,
                                  args=[obj.name])


class GroupsColumn(Column):
    """Shows the list of groups requested to review the review request."""

    label = _('Groups')
    detailed_label = _('Target Groups')
    sortable = False
    shrink = False

    def augment_queryset(
        self,
        state: StatefulColumn,
        queryset: QuerySet,
    ) -> QuerySet:
        """Add additional queries to the queryset.

        This will pre-fetch the review groups for each review request.

        Args:
            state (djblets.datagrid.grids.StatefulColumn):
                The column state.

            queryset (django.db.models.query.QuerySet):
                The queryset to augment.

        Returns:
            django.db.models.query.QuerySet:
            The resulting queryset.
        """
        return queryset.prefetch_related('target_groups')

    def render_data(
        self,
        state: StatefulColumn,
        obj: ReviewRequest,
    ) -> SafeString:
        """Return the rendered contents of the column.

        Args:
            state (djblets.datagrid.grids.StatefulColumn):
                The state for the datagrid.

            obj (reviewboard.reviews.models.review_request.ReviewRequest):
                The review request.

        Returns:
            django.utils.safestring.SafeString:
            The rendered HTML.
        """
        return escape(' '.join(
            group.name
            for group in self.get_raw_object_value(state, obj)
        ))

    def get_raw_object_value(
        self,
        state: StatefulColumn,
        obj: ReviewRequest,
    ) -> Sequence[Group]:
        """Return the list of review groups.

        Version Added:
            8.0

        Args:
            state (djblets.datagrids.grids.StatefulColumn):
                The state for the DataGrid instance.

            obj (reviewboard.reviews.models.ReviewRequest):
                The review request to return the value for.

        Returns:
            list of reviewboard.reviews.models.Group:
            The list of review groups.
        """
        return list(obj.target_groups.all())


class MyCommentsColumn(Column):
    """Shows if the current user has reviewed the review request."""

    detailed_label = _('My Comments')
    image_alt = _('My Comments')
    image_class = 'rb-icon rb-icon-datagrid-comment-draft'
    shrink = True

    #: A mapping of computed statuses to icon CSS classes.
    #:
    #: Version Added:
    #:     8.0
    _ICON_CLASSES: Final[Mapping[str, str]] = {
        'has-draft-review': 'rb-icon-datagrid-comment-draft',
        'has-review': 'rb-icon-datagrid-comment',
        'has-ship-it': 'rb-icon-datagrid-comment-shipit',
    }

    #: A mapping of computed statuses to image alt text.
    #:
    #: Version Added:
    #:     8.0
    _IMAGE_ALTS: Final[Mapping[str, StrPromise]] = {
        'has-draft-review': _('Comments drafted'),
        'has-review': _('Comments published'),
        'has-ship-it': _('Comments published. Ship it!'),
    }

    def augment_queryset(
        self,
        state: StatefulColumn,
        queryset: QuerySet,
    ) -> QuerySet:
        """Add additional queries to the queryset.

        This will add subqueries for determining whether there are
        posted reviews, draft reviews, or Ship It! reviews.

        Args:
            state (djblets.datagrid.grids.StatefulColumn):
                The state for the DataGrid instance.

            queryset (django.db.models.query.QuerySet):
                The queryset to augment.

        Returns:
            django.db.models.query.QuerySet:
            The resulting augmented QuerySet.
        """
        user = state.datagrid.request.user

        if user.is_anonymous:
            return queryset

        query_dict = {
            'user_id': str(user.pk),
        }

        return queryset.extra(select={
            'mycomments_my_reviews': """
                SELECT COUNT(1)
                  FROM reviews_review
                  WHERE reviews_review.user_id = %(user_id)s
                    AND reviews_review.review_request_id =
                        reviews_reviewrequest.id
            """ % query_dict,
            'mycomments_private_reviews': """
                SELECT COUNT(1)
                  FROM reviews_review
                  WHERE reviews_review.user_id = %(user_id)s
                    AND reviews_review.review_request_id =
                        reviews_reviewrequest.id
                    AND NOT reviews_review.public
            """ % query_dict,
            'mycomments_shipit_reviews': """
                SELECT COUNT(1)
                  FROM reviews_review
                  WHERE reviews_review.user_id = %(user_id)s
                    AND reviews_review.review_request_id =
                        reviews_reviewrequest.id
                    AND reviews_review.ship_it
            """ % query_dict,
        })

    def render_data(
        self,
        state: StatefulColumn,
        obj: ReviewRequest,
    ) -> SafeString:
        """Return the rendered contents of the column.

        Args:
            state (djblets.datagrids.grids.StatefulColumn):
                The state for the DataGrid instance.

            obj (reviewboard.reviews.models.ReviewRequest):
                The review request being rendered for this row.
        """
        info = self.get_raw_object_value(state, obj)

        if not info:
            return mark_safe('')

        status = info['status']

        return format_html(
            '<div class="rb-icon {}" title="{}"></div>',
            self._ICON_CLASSES[status],
            self._IMAGE_ALTS[status],
        )

    def get_raw_object_value(
        self,
        state: StatefulColumn,
        obj: ReviewRequest,
    ) -> Mapping[str, Any]:
        """Return the raw value for review status.

        Version Added:
            8.0

        Args:
            state (djblets.datagrids.grids.StatefulColumn):
                The state for the DataGrid instance.

            obj (reviewboard.reviews.models.ReviewRequest):
                The review request for the row.

        Returns:
            dict:
            Information on the review status.
        """
        user = state.datagrid.request.user

        if user.is_anonymous or obj.mycomments_my_reviews == 0:
            return {}

        has_draft_reviews = bool(getattr(obj, 'mycomments_private_reviews'))
        has_shipit_reviews = bool(getattr(obj, 'mycomments_shipit_reviews'))
        status: str = ''

        # Priority is ranked in the following order:
        #
        # 1) Non-public (draft) reviews
        # 2) Public reviews marked "Ship It"
        # 3) Public reviews not marked "Ship It"
        if has_draft_reviews:
            status = 'has-draft-review'
        elif has_shipit_reviews:
            status = 'has-ship-it'
        else:
            status = 'has-review'

        return {
            'has_draft_reviews': has_draft_reviews,
            'has_ship_its': has_shipit_reviews,
            'status': status,
        }


class NewUpdatesColumn(Column):
    """Indicates if there are new updates on a review request.

    This will show an icon if the review request has had any new updates
    or reviews since the user last saw it.
    """

    detailed_label = _('New Updates')
    image_alt = _('New Updates')
    image_class = 'rb-icon rb-icon-new-updates'
    shrink = True

    def render_data(
        self,
        state: StatefulColumn,
        obj: ReviewRequest,
    ) -> SafeString:
        """Return the rendered contents of the column.

        Args:
            state (djblets.datagrids.grids.StatefulColumn):
                The state for the DataGrid instance.

            obj (reviewboard.reviews.models.ReviewRequest):
                The group or review request being rendered for this row.
        """
        new_review_count = self.get_raw_object_value(state, obj)

        if new_review_count > 0:
            return format_html(
                '<div class="{}" title="{}"></div>',
                self.image_class, self.image_alt,
            )

        return mark_safe('')

    def get_raw_object_value(
        self,
        state: StatefulColumn,
        obj: ReviewRequest,
    ) -> int:
        """Return the raw value for the new review count.

        Version Added:
            8.0

        Args:
            state (djblets.datagrids.grids.StatefulColumn):
                The state for the DataGrid instance.

            obj (reviewboard.reviews.models.ReviewRequest):
                The review request for the row.

        Returns:
            int:
            The new review count for the review request.
        """
        # Review requests for un-authenticated users will not contain the
        # new_review_count attribute.
        return getattr(obj, 'new_review_count', 0)


class PendingCountColumn(Column):
    """Shows the pending number of review requests for a user or group.

    This will show the pending number of review requests for the given
    review group or user. It only applies to group or user lists.
    """

    def augment_queryset_for_data(
        self,
        state: StatefulColumn,
        queryset: QuerySet,
        **kwargs,
    ) -> QuerySet:
        """Augment a queryset for data-rendering purposes.

        This will count the number of accessible open review requests.

        Version Added:
            5.0.7

        Args:
            state (djblets.datagrid.grids.StatefulColumn):
                The state for the DataGrid instance.

            queryset (django.db.models.query.QuerySet):
                The queryset to augment.

            **kwargs (dict):
                Additional keyword arguments for future expansion.

        Returns:
            django.db.models.query.QuerySet:
            The resulting augmented QuerySet.
        """
        return queryset.annotate(
            column_pending_review_request_count=Count(
                'review_requests',
                filter=(Q(review_requests__public=True) &
                        Q(review_requests__status='P'))))

    def render_data(
        self,
        state: StatefulColumn,
        obj: Group,
    ) -> SafeString:
        """Return the rendered contents of the column.

        Args:
            state (djblets.datagrid.grids.StatefulColumn):
                The state for the datagrid.

            review_request (reviewboard.reviews.models.review_request.
                            ReviewRequest):
                The review request.

        Returns:
            django.utils.safestring.SafeString:
            The rendered HTML.
        """
        return escape(str(self.get_raw_object_value(state, obj)))

    def get_raw_object_value(
        self,
        state: StatefulColumn,
        obj: Group,
    ) -> int:
        """Return the raw value for the pending review request count.

        Version Added:
            8.0

        Args:
            state (djblets.datagrids.grids.StatefulColumn):
                The state for the DataGrid instance.

            obj (reviewboard.reviews.models.Group):
                The review request for the row.

        Returns:
            int:
            The pending review request count.
        """
        return getattr(obj, 'column_pending_review_request_count', 0)


class PeopleColumn(Column):
    """Shows the list of people requested to review the review request."""

    label = _('People')
    detailed_label = _('Target People')
    sortable = False
    shrink = False

    def augment_queryset(
        self,
        state: StatefulColumn,
        queryset: QuerySet,
    ) -> QuerySet:
        """Add additional queries to the queryset.

        This will pre-fetch the reviewers for each review request.

        Args:
            state (djblets.datagrid.grids.StatefulColumn):
                The column state.

            queryset (django.db.models.query.QuerySet):
                The queryset to augment.

        Returns:
            django.db.models.query.QuerySet:
            The resulting queryset.
        """
        return queryset.prefetch_related('target_people')

    def render_data(
        self,
        state: StatefulColumn,
        obj: ReviewRequest,
    ) -> SafeString:
        """Return the rendered contents of the column.

        Args:
            state (djblets.datagrid.grids.StatefulColumn):
                The state for the datagrid.

            obj (reviewboard.reviews.models.review_request.ReviewRequest):
                The review request.

        Returns:
            django.utils.safestring.SafeString:
            The rendered HTML.
        """
        return escape(' '.join(
            user.username
            for user in self.get_raw_object_value(state, obj)
        ))

    def get_raw_object_value(
        self,
        state: StatefulColumn,
        obj: ReviewRequest,
    ) -> Sequence[User]:
        """Return the list of reviewers.

        Version Added:
            8.0

        Args:
            state (djblets.datagrids.grids.StatefulColumn):
                The state for the DataGrid instance.

            obj (reviewboard.reviews.models.ReviewRequest):
                The review request to return the value for.

        Returns:
            list of django.contrib.auth.models.User:
            The list of reviewers.
        """
        return list(obj.target_people.all())


class RepositoryColumn(Column):
    """Shows the name of the repository the review request's change is on."""

    field_name = 'repository'
    label = _('Repository')
    db_field = 'repository__name'
    css_class = 'repository-column'
    shrink = True
    sortable = True
    link = False

    def augment_queryset(
        self,
        state: StatefulColumn,
        queryset: QuerySet,
    ) -> QuerySet:
        """Add additional queries to the queryset.

        This will include the repository for each review request.

        Args:
            state (djblets.datagrid.grids.StatefulColumn):
                The column state.

            queryset (django.db.models.query.QuerySet):
                The queryset to augment.

        Returns:
            django.db.models.query.QuerySet:
            The resulting queryset.
        """
        return queryset.select_related('repository')

    def render_data(
        self,
        state: StatefulColumn,
        obj: ReviewRequest,
    ) -> SafeString:
        """Return the rendered contents of the column.

        Args:
            state (djblets.datagrid.grids.StatefulColumn):
                The state for the datagrid.

            obj (reviewboard.reviews.models.review_request.ReviewRequest):
                The review request.

        Returns:
            django.utils.safestring.SafeString:
            The rendered HTML.
        """
        if obj.repository_id is None:
            return mark_safe('')

        return escape(obj.repository.name)


class ReviewCountColumn(Column):
    """Shows the number of published reviews for a review request."""

    label = _('Reviews')
    detailed_label = _('Number of Reviews')
    shrink = True
    link = True

    def augment_queryset(
        self,
        state: StatefulColumn,
        queryset: QuerySet,
    ) -> QuerySet:
        """Add additional queries to the queryset.

        This will perform a subquery to return a public review count for
        each review request.

        Args:
            state (djblets.datagrid.grids.StatefulColumn):
                The column state.

            queryset (django.db.models.query.QuerySet):
                The queryset to augment.

        Returns:
            django.db.models.query.QuerySet:
            The resulting queryset.
        """
        return queryset.extra(select={
            'publicreviewcount_count': """
                SELECT COUNT(*)
                  FROM reviews_review
                  WHERE reviews_review.public
                    AND reviews_review.base_reply_to_id is NULL
                    AND reviews_review.review_request_id =
                        reviews_reviewrequest.id
            """
        })

    def render_data(
        self,
        state: StatefulColumn,
        obj: ReviewRequest,
    ) -> SafeString:
        """Return the rendered contents of the column.

        Args:
            state (djblets.datagrid.grids.StatefulColumn):
                The state for the datagrid.

            obj (reviewboard.reviews.models.review_request.ReviewRequest):
                The review request.

        Returns:
            django.utils.safestring.SafeString:
            The rendered HTML for the column.
        """
        return escape(str(self.get_raw_object_value(state, obj)))

    def get_raw_object_value(
        self,
        state: StatefulColumn,
        obj: ReviewRequest,
    ) -> int:
        """Return the count value for the review request.

        Version Added:
            8.0

        Args:
            state (djblets.datagrids.grids.StatefulColumn):
                The state for the DataGrid instance.

            obj (reviewboard.reviews.models.ReviewRequest):
                The review request to return the value for.

        Returns:
            int:
            The review count value.
        """
        return getattr(obj, 'publicreviewcount_count')

    @staticmethod
    def link_func(
        state: StatefulColumn,
        obj: ReviewRequest,
        value: str,
    ) -> str:
        """Return the link to the object in the column.

        Args:
            state (djblets.datagrid.grids.StatefulColumn):
                The column state.

            obj (reviewboard.reviews.models.ReviewRequest):
                The review request for the row.

            value (str):
                The rendered data.

        Returns:
            str:
            The URL to link to.
        """
        return '%s#last-review' % obj.get_absolute_url()


class ReviewGroupStarColumn(BaseStarColumn):
    """Indicates if a review group is starred.

    The star is interactive, allowing the user to star or unstar the group.
    """

    starrable_model = Group


class ReviewRequestIDColumn(Column):
    """Displays the ID of the review request."""

    label = _('ID')
    detailed_label = _('Review Request ID')
    field_name = 'display_id'
    shrink = True
    sortable = True
    link = True

    def get_sort_field(
        self,
        state: StatefulColumn,
    ) -> str:
        """Return the model field for sorting this column.

        Args:
            state (djblets.datagrids.grids.StatefulColumn):
                The state for the DataGrid instance.

        Returns:
            str:
            The sort field.
        """
        if state.datagrid.local_site:
            return 'local_id'
        else:
            return 'id'


class ReviewRequestStarColumn(BaseStarColumn):
    """Indicates if a review request is starred.

    The star is interactive, allowing the user to star or unstar the
    review request.
    """

    starrable_model = ReviewRequest


class ShipItColumn(Column):
    """Shows the "Ship It" and issue counts for a review request.

    If there are any issues still to resolve or verify, this will instead show
    information on those issues. Otherwise, it will show information on the
    number of Ship It! reviews filed.

    The following is the order of priority in which information is shown:

    1. Open issues with issues requiring verification
    2. Open issues
    3. Issues requiring verification
    4. Ship It! counts

    If showing a Ship It!, and if the latest review is older than the last
    update on the review request, the Ship It! will be marked as stale,
    helping visually indicate that it may need a re-review. The ARIA label
    reflects this as well.

    Version Changed:
        5.0:
        * Added ARIA attributes for the displayed output.
        * Ship It! counts are now shown as stale if older than the latest
          update to the review request.
    """

    detailed_label = _('Ship It!/Issue Counts')
    db_field = 'shipit_count'
    image_alt = _('Ship It!/Issue Counts')
    image_class = 'rb-icon rb-icon-datagrid-column-shipits-issues'
    shrink = True
    sortable = True

    def render_data(
        self,
        state: StatefulColumn,
        obj: ReviewRequest,
    ) -> SafeString:
        """Return the rendered contents of the column.

        Args:
            state (djblets.datagrid.grids.StatefulColumn):
                The state for the datagrid.

            obj (reviewboard.reviews.models.review_request.ReviewRequest):
                The review request.

        Returns:
            django.utils.safestring.SafeString:
            The rendered HTML for the column.
        """
        info = self.get_raw_object_value(state, obj)

        open_issues = info['issue_open_count']
        verifying_issues = info['issue_verifying_count']
        shipit_count = info['shipit_count']

        if open_issues > 0 and verifying_issues > 0:
            return self._render_counts(
                [
                    {
                        'count': open_issues,
                        'title': _('Open issue count'),
                    },
                    {
                        'count': verifying_issues,
                        'css_class': 'issue-verifying-count',
                        'icon_name': 'issue-verifying',
                        'title': _('Verifying issue count'),
                    },
                ],
                aria_label=(
                    ngettext(
                        '%(open_issue_count)d issue opened, '
                        '%(verifying_issue_count)d requiring verification',
                        '%(open_issue_count)d issues opened, '
                        '%(verifying_issue_count)d requiring verification',
                        open_issues)
                    % {
                        'open_issue_count': open_issues,
                        'verifying_issue_count': verifying_issues,
                    }
                ))
        elif open_issues > 0:
            return self._render_counts(
                [{
                    'count': open_issues,
                }],
                aria_label=(
                    ngettext(
                        '%(open_issue_count)d issue opened',
                        '%(open_issue_count)d issues opened',
                        open_issues)
                    % {
                        'open_issue_count': open_issues,
                    }
                ))
        elif verifying_issues > 0:
            return self._render_counts(
                [{
                    'count': verifying_issues,
                    'icon_name': 'issue-verifying',
                }],
                aria_label=(
                    ngettext(
                        '%(verifying_issue_count)d issue requiring '
                        'verification',
                        '%(verifying_issue_count)d issues requiring '
                        'verification',
                        verifying_issues)
                    % {
                        'verifying_issue_count': verifying_issues,
                    }
                ))
        elif shipit_count > 0:
            container_css_classes = ['shipit-count-container']

            if info.get('shipits_stale'):
                container_css_classes.append('-is-stale')
                aria_label = ngettext(
                    "%(shipit_count)d Ship It! (New updates to review)",
                    "%(shipit_count)d Ship It's! (New updates to review)",
                    shipit_count)
            else:
                aria_label = ngettext(
                    "%(shipit_count)d Ship It!",
                    "%(shipit_count)d Ship It's!",
                    shipit_count)

            return self._render_counts(
                [{
                    'count': shipit_count,
                    'css_class': 'shipit-count',
                    'icon_name': 'shipit',
                }],
                aria_label=aria_label % {
                    'shipit_count': shipit_count,
                },
                container_css_class=' '.join(container_css_classes))
        else:
            return mark_safe('')

    def get_raw_object_value(
        self,
        state: StatefulColumn,
        obj: ReviewRequest,
    ) -> Mapping[str, Any]:
        """Return the summary and labels for a review request.

        Version Added:
            8.0

        Args:
            state (djblets.datagrids.grids.StatefulColumn):
                The state for the DataGrid instance.

            obj (reviewboard.reviews.models.ReviewRequest):
                The review request to calculate summary information for.

        Returns:
            dict:
            The summary and labels for the review request.
        """
        review_request = obj
        open_issues = review_request.issue_open_count
        verifying_issues = review_request.issue_verifying_count
        shipit_count = review_request.shipit_count

        return {
            'issue_open_count': open_issues,
            'issue_verifying_count': verifying_issues,
            'shipit_count': shipit_count,
            'shipits_stale': bool(
                review_request.last_review_activity_timestamp and
                review_request.last_updated and
                (review_request.last_review_activity_timestamp <
                 review_request.last_updated)
            ),
        }

    def _render_counts(
        self,
        count_details: Sequence[Mapping[str, Any]],
        aria_label: str,
        container_css_class: str = 'issue-count-container',
    ) -> SafeString:
        """Render the counts for the column.

        This will render a container bubble in the column and render each
        provided count and icon in the bubble. This can be used for issues,
        Ship Its, or anything else we need down the road.

        Version Added:
            5.0:
            * Added ``aria_label``.
            * Removed the ``title`` key from ``count_details``.

        Args:
            count_details (list of dict):
                The list of details for the count. Each supports the following
                keys:

                Keys:
                    count (int):
                        The count to show for the indicator.

                    css_class (str, optional):
                        The CSS class to use for the indicator.

                    icon_name (str, optional):
                        The name of the icon to use for the indicator.

            aria_label (str):
                The label to use for the ``aria-label`` attribute and the
                ``title`` attribute.

            container_css_class (str, optional):
                The optional CSS class name for the outer container.

        Returns:
            django.utils.safestring.SafeString:
            The resulting HTML for the counts bubble.
        """
        # Note that the HTML is very whitespace-sensitive, so don't try to
        # change the templates to be nicely indented. The spacing is this way
        # for a reason.
        #
        # We also can't use format_html_join, unfortunately, as that doesn't
        # support keyword arguments.
        return format_html(
            '<div class="{container_css_class}" aria-label="{aria_label}"'
            ' title="{aria_label}">'
            '{count_html}'
            '</div>',
            aria_label=aria_label,
            container_css_class=container_css_class,
            count_html=mark_safe(''.join(
                format_html(
                    '<span class="{css_class}" aria-hidden="true">'
                    '<span class="rb-icon rb-icon-datagrid-{icon_name}">'
                    '</span>'
                    '{count}'
                    '</span>',
                    **{
                        'css_class': 'issue-count',
                        'icon_name': 'open-issues',
                        **count_detail,
                    },
                )
                for count_detail in count_details
            )))


class SummaryColumn(Column):
    """Shows the summary of a review request.

    This will also prepend the draft/submitted/discarded state, if any,
    to the summary.
    """

    label = _('Summary')
    css_class = 'summary'
    expand = True
    sortable = True
    link = True
    link_css_class = 'review-request-link'

    def augment_queryset(
        self,
        state: StatefulColumn,
        queryset: QuerySet,
    ) -> QuerySet:
        """Add additional queries to the queryset.

        This will perform subqueries to retrieve draft summaries and
        archive/mute states for the review requests, in order to show
        labels before the summaries.

        Args:
            state (djblets.datagrid.grids.StatefulColumn):
                The column state.

            queryset (django.db.models.query.QuerySet):
                The queryset to augment.

        Returns:
            django.db.models.query.QuerySet:
            The resulting queryset.
        """
        user = state.datagrid.request.user

        if user.is_anonymous:
            return queryset

        return queryset.extra(select={
            'draft_summary': """
                SELECT reviews_reviewrequestdraft.summary
                  FROM reviews_reviewrequestdraft
                  WHERE reviews_reviewrequestdraft.review_request_id =
                        reviews_reviewrequest.id
            """,
            'visibility': """
                SELECT accounts_reviewrequestvisit.visibility
                  FROM accounts_reviewrequestvisit
                 WHERE accounts_reviewrequestvisit.review_request_id =
                       reviews_reviewrequest.id
                   AND accounts_reviewrequestvisit.user_id = %(user_id)s
            """ % {
                'user_id': str(user.pk),
            }
        })

    def render_data(
        self,
        state: StatefulColumn,
        obj: ReviewRequest,
    ) -> SafeString:
        """Return the rendered contents of the column.

        Args:
            state (djblets.datagrid.grids.StatefulColumn):
                The state for the datagrid.

            obj (reviewboard.reviews.models.review_request.ReviewRequest):
                The review request.

        Returns:
            django.utils.safestring.SafeString:
            The rendered column.
        """
        info = self.get_raw_object_value(state, obj)
        summary = info['summary']
        labels = info['labels']

        label_status_map = {
            'completed': 'submitted',
        }

        result: list[str] = []

        for label_info in labels:
            status = label_info['status']

            result.append(format_html(
                '<label class="label-{}">{}</label>',
                label_status_map.get(status, status),
                label_info['label'],
            ))

        if summary:
            result.append(format_html('<span>{}</span>', summary))
        else:
            result.append(format_html('<span class="no-summary">{}</span>',
                                      _('No Summary')))

        return mark_safe(''.join(result))

    def get_raw_object_value(
        self,
        state: StatefulColumn,
        obj: ReviewRequest,
    ) -> Mapping[str, Any]:
        """Return the summary and labels for a review request.

        Version Added:
            8.0

        Args:
            state (djblets.datagrids.grids.StatefulColumn):
                The state for the DataGrid instance.

            obj (reviewboard.reviews.models.ReviewRequest):
                The review request to calculate summary information for.

        Returns:
            dict:
            The summary and labels for the review request.
        """
        review_request = obj
        summary = review_request.summary
        labels: list[Mapping[str, str]] = []

        if review_request.submitter_id == state.datagrid.request.user.id:
            draft_summary = getattr(review_request, 'draft_summary')

            if draft_summary is not None:
                summary = draft_summary
                labels.append({
                    'label': gettext('Draft'),
                    'status': 'draft',
                })
            elif (not review_request.public and
                  review_request.status == ReviewRequest.PENDING_REVIEW):
                labels.append({
                    'label': gettext('Draft'),
                    'status': 'draft',
                })

        # review_request.visibility is not defined when the user is not
        # logged in.
        if state.datagrid.request.user.is_authenticated:
            visibility = getattr(review_request, 'visibility')

            if visibility == ReviewRequestVisit.ARCHIVED:
                labels.append({
                    'label': gettext('Archived'),
                    'status': 'archived',
                })
            elif visibility == ReviewRequestVisit.MUTED:
                labels.append({
                    'label': gettext('Muted'),
                    'status': 'muted',
                })

        if review_request.status == ReviewRequest.SUBMITTED:
            labels.append({
                'label': gettext('Completed'),
                'status': 'completed',
            })
        elif review_request.status == ReviewRequest.DISCARDED:
            labels.append({
                'label': gettext('Discarded'),
                'status': 'discarded',
            })

        return {
            'labels': labels,
            'summary': summary,
        }


class ReviewSummaryColumn(Column):
    """Shows the summary of the review request of a review.

    This does not (yet) prepend the draft/submitted/discarded state, if any,
    to the summary.
    """

    label = _('Review Request Summary')
    cell_template = 'datagrids/review_summary_cell.html'
    css_class = 'summary'
    expand = True
    link = True
    link_css_class = 'review-request-link'

    def augment_queryset(
        self,
        state: StatefulColumn,
        queryset: QuerySet,
    ) -> QuerySet:
        """Add additional queries to the queryset.

        This will load the review requests along with each review.

        Args:
            state (djblets.datagrid.grids.StatefulColumn):
                The column state.

            queryset (django.db.models.query.QuerySet):
                The queryset to augment.

        Returns:
            django.db.models.query.QuerySet:
            The resulting queryset.
        """
        return queryset.select_related('review_request')

    def get_raw_object_value(
        self,
        state: StatefulColumn,
        obj: Review,
    ) -> str:
        """Return the raw value for review summary.

        This will fetch the summary from the associated review request.

        Version Added:
            8.0

        Args:
            state (djblets.datagrids.grids.StatefulColumn):
                The state for the DataGrid instance.

            obj (reviewboard.reviews.models.Review):
                The review being rendered.

        Returns:
            str:
            The summary of the associated review request.
        """
        return obj.review_request.summary

    def render_cell(
        self,
        state: StatefulColumn,
        obj: Review,
        render_context: _RenderContext,
    ) -> SafeString:
        """Render the table cell containing column data.

        Version Added:
            8.0

        Args:
            state (djblets.datagrid.grids.StatefulColumn):
                The state for the DataGrid instance.

            obj (reviewboard.reviews.models.Review):
                The review being rendered.

            render_context (dict or django.template.context.Context):
                The shared context used for cell renders.

        Returns:
            django.utils.safestring.SafeString:
            The rendered cell as HTML.
        """
        # We pass this to the cell's review request info box.
        render_context['review_request_url'] = \
            obj.review_request.get_absolute_url()

        try:
            return super().render_cell(state, obj, render_context)
        finally:
            del render_context['review_request_url']


class ToMeColumn(Column):
    """Indicates if the user is requested to review the change.

    This will show an indicator if the user is on the Target People reviewers
    list.
    """

    label = '\u00BB'
    detailed_label = _('To Me')
    detailed_label_html = _('\u00BB To Me')
    shrink = True

    def augment_queryset(
        self,
        state: StatefulColumn,
        queryset: QuerySet,
    ) -> QuerySet:
        """Add additional queries to the queryset.

        This will load the PKs of the review requests directed to the
        user and store them as state on the column for rendering.

        Args:
            state (djblets.datagrid.grids.StatefulColumn):
                The column state.

            queryset (django.db.models.query.QuerySet):
                The queryset to augment.

        Returns:
            django.db.models.query.QuerySet:
            The resulting queryset.
        """
        user = state.datagrid.request.user

        if user.is_authenticated:
            pks = set(
                user.directed_review_requests
                .filter(pk__in=state.datagrid.id_list)
                .values_list('pk', flat=True)
            )
        else:
            pks = set()

        state.extra_data['all_to_me'] = pks

        return queryset

    def render_data(
        self,
        state: StatefulColumn,
        obj: ReviewRequest,
    ) -> SafeString:
        """Return the rendered contents of the column.

        If the object was directed to the user, an indicator will be
        shown.

        Args:
            state (djblets.datagrids.grids.StatefulColumn):
                The state for the DataGrid instance.

            obj (reviewboard.reviews.models.ReviewRequest):
                The review request the column applies to.

        Returns:
            django.utils.safestring.SafeString:
            The rendered HTML.
        """
        if self.get_raw_object_value(state, obj):
            return format_html(
                '<div title="{}"><b>&raquo;</b></div>',
                self.detailed_label,
            )

        return mark_safe('')

    def get_raw_object_value(
        self,
        state: StatefulColumn,
        obj: ReviewRequest,
    ) -> bool:
        """Return the "to me" state as a boolean.

        Version Added:
            8.0

        Args:
            state (djblets.datagrids.grids.StatefulColumn):
                The state for the DataGrid instance.

            obj (reviewboard.reviews.models.ReviewRequest):
                The review request for the row.

        Returns:
            bool:
            ``True`` if the column was directed to the user. ``False`` if
            it was not.
        """
        return obj.pk in state.extra_data['all_to_me']


class DiffSizeColumn(Column):
    """Indicates line add/delete counts for the latest diffset."""

    label = _('Diff Size')
    shrink = True

    def render_data(
        self,
        state: StatefulColumn,
        obj: ReviewRequest,
    ) -> SafeString:
        """Return the rendered contents of the column.

        This will format HTML showing the diff's insert and delete
        line counts.

        Args:
            state (djblets.datagrids.grids.StatefulColumn):
                The state for the DataGrid instance.

            obj (reviewboard.reviews.models.ReviewRequest):
                The review request the column applies to.

        Returns:
            django.utils.safestring.SafeString:
            The rendered HTML.
        """
        counts = self.get_raw_object_value(state, obj)

        if counts:
            insert_count = counts.get('raw_insert_count')
            delete_count = counts.get('raw_delete_count')
            result: list[SafeString] = []

            if insert_count:
                result.append(format_html(
                    '<span class="diff-size-column insert">+{}</span>',
                    insert_count,
                ))

            if delete_count:
                result.append(format_html(
                    '<span class="diff-size-column delete">-{}</span>',
                    delete_count,
                ))

            if result:
                return mark_safe('&nbsp;'.join(result))

        return mark_safe('')

    def get_raw_object_value(
        self,
        state: StatefulColumn,
        obj: ReviewRequest,
    ) -> Mapping[str, int] | None:
        """Return the raw value for the diff size information.

        This will return the line count information for the diffset.

        Version Added:
            8.0

        Args:
            state (djblets.datagrids.grids.StatefulColumn):
                The state for the DataGrid instance.

            obj (reviewboard.reviews.models.ReviewRequest):
                The review request for the row.

        Returns:
            dict:
            The list of diff size information.
        """
        if obj.repository_id is not None:
            diffset_history = obj.diffset_history
            assert diffset_history is not None

            diffsets = list(diffset_history.diffsets.all())

            if diffsets:
                return diffsets[-1].get_total_line_counts()

        return None

    def augment_queryset(
        self,
        state: StatefulColumn,
        queryset: QuerySet,
    ) -> QuerySet:
        """Add additional queries to the queryset.

        This will prefetch the diffsets and filediffs needed to perform the
        line calculations.

        Args:
            state (djblets.datagrid.grids.StatefulColumn):
                The column state.

            queryset (django.db.models.query.QuerySet):
                The queryset to augment.

        Returns:
            django.db.models.query.QuerySet:
            The resulting queryset.
        """
        # TODO: Update this to fetch only the specific fields when we move
        #       to a newer version of Django.
        return queryset.prefetch_related('diffset_history__diffsets',
                                         'diffset_history__diffsets__files')



"""Unit tests for reviewboard.datagrids.columns.DiffSizeColumn.

Version Added:
    8.0
"""

from __future__ import annotations

from django.utils.safestring import SafeString

from reviewboard.datagrids.columns import DiffSizeColumn
from reviewboard.datagrids.tests.base import BaseColumnTestCase


class DiffSizeColumnTests(BaseColumnTestCase):
    """Unit tests for reviewboard.datagrids.columns.DiffSizeColumn.

    Version Added:
        8.0
    """

    column = DiffSizeColumn()
    fixtures = ['test_users', 'test_scmtools']

    def test_render_data_with_no_repository(self) -> None:
        """Testing DiffSizeColumn.render_data with no repository"""
        review_request = self.create_review_request(publish=True)

        value = self.column.render_data(self.stateful_column, review_request)

        self.assertIsInstance(value, SafeString)
        self.assertEqual(value, '')

    def test_render_data_with_repository_no_diff(self) -> None:
        """Testing DiffSizeColumn.render_data with repository but no diff"""
        review_request = self.create_review_request(
            create_repository=True,
            publish=True)

        value = self.column.render_data(self.stateful_column, review_request)

        self.assertIsInstance(value, SafeString)
        self.assertEqual(value, '')

    def test_render_data_with_inserts_and_deletes(self) -> None:
        """Testing DiffSizeColumn.render_data with inserts and deletes"""
        review_request = self.create_review_request(
            create_repository=True,
            publish=True)
        diffset = self.create_diffset(review_request)
        self.create_filediff(
            diffset,
            extra_data={
                'raw_insert_count': 10,
                'raw_delete_count': 5,
            })

        value = self.column.render_data(self.stateful_column, review_request)

        self.assertIsInstance(value, SafeString)
        self.assertHTMLEqual(
            value,
            '<span class="diff-size-column insert">+10</span>&nbsp;'
            '<span class="diff-size-column delete">-5</span>')

    def test_render_data_with_inserts_only(self) -> None:
        """Testing DiffSizeColumn.render_data with inserts only"""
        review_request = self.create_review_request(
            create_repository=True,
            publish=True)
        diffset = self.create_diffset(review_request)
        self.create_filediff(
            diffset,
            extra_data={
                'raw_insert_count': 7,
                'raw_delete_count': 0,
            })

        value = self.column.render_data(self.stateful_column, review_request)

        self.assertIsInstance(value, SafeString)
        self.assertHTMLEqual(
            value,
            '<span class="diff-size-column insert">+7</span>')

    def test_render_data_with_deletes_only(self) -> None:
        """Testing DiffSizeColumn.render_data with deletes only"""
        review_request = self.create_review_request(
            create_repository=True,
            publish=True)
        diffset = self.create_diffset(review_request)
        self.create_filediff(
            diffset,
            extra_data={
                'raw_insert_count': 0,
                'raw_delete_count': 3,
            })

        value = self.column.render_data(self.stateful_column, review_request)

        self.assertIsInstance(value, SafeString)
        self.assertHTMLEqual(
            value,
            '<span class="diff-size-column delete">-3</span>')

    def test_to_json_with_no_repository(self) -> None:
        """Testing DiffSizeColumn.to_json with no repository"""
        review_request = self.create_review_request(publish=True)

        self.assertIsNone(self.column.to_json(self.stateful_column,
                                              review_request))

    def test_to_json_with_diff(self) -> None:
        """Testing DiffSizeColumn.to_json with diff"""
        review_request = self.create_review_request(
            create_repository=True,
            publish=True)
        diffset = self.create_diffset(review_request)
        self.create_filediff(
            diffset,
            extra_data={
                'raw_insert_count': 10,
                'raw_delete_count': 5,
            })

        self.assertEqual(
            self.column.to_json(self.stateful_column, review_request),
            {
                'delete_count': 5,
                'equal_count': None,
                'insert_count': 10,
                'raw_delete_count': 5,
                'raw_insert_count': 10,
                'replace_count': None,
                'total_line_count': None,
            })



"""Management command for disabling an extension."""

from django.core.management.base import BaseCommand, CommandError
from django.utils.translation import gettext as _

from reviewboard.extensions.base import get_extension_manager


class Command(BaseCommand):
    """Management command for disabling an extension."""

    help = _('Disables an extension.')

    def add_arguments(self, parser):
        """Add arguments to the command.

        Args:
            parser (argparse.ArgumentParser):
                The argument parser for the command.
        """
        parser.add_argument(
            'extension_ids',
            metavar='EXTENSION_ID',
            nargs='*',
            help=_('The ID of the extension to disable.'))

    def handle(self, *args, **options):
        """Handle the command.

        Args:
            *args (tuple):
                The name of the check to resolve.

            **options (dict, unused):
                Options parsed on the command line. For this command, no
                options are available.

        Raises:
            django.core.management.CommandError:
                There was an error with arguments or enabling the extension.
        """
        extension_ids = options['extension_ids']

        if not extension_ids:
            raise CommandError(
                _('You must specify an extension ID to disable.'))

        extension_mgr = get_extension_manager()

        for extension_id in extension_ids:
            try:
                extension_mgr.disable_extension(extension_id)
            except Exception as e:
                raise CommandError(
                    _('Unexpected error disabling extension %s: %s')
                    % (extension_id, e))



"""Convert Repository.bug_tracker from a URLField to a CharField.

Version Added:
    1.0
"""

from django_evolution.mutations import ChangeField


MUTATIONS = [
    ChangeField('Repository', 'bug_tracker', max_length=256),
]



"""Views for interacting with bug trackers."""

from __future__ import annotations

import re
from typing import TYPE_CHECKING

from django.http import (HttpRequest,
                         HttpResponse,
                         HttpResponseNotFound)
from django.utils.html import escape, strip_tags
from django.utils.safestring import SafeString, mark_safe
from django.utils.translation import gettext_lazy as _
from django.views.generic.base import TemplateView, View

from reviewboard.hostingsvcs.base.bug_tracker import BaseBugTracker
from reviewboard.reviews.markdown_utils import render_markdown
from reviewboard.reviews.views.mixins import ReviewRequestViewMixin
from reviewboard.site.urlresolvers import local_site_reverse

if TYPE_CHECKING:
    from typing import Any


class BugInfoboxView(ReviewRequestViewMixin, TemplateView):
    """Displays information on a bug, for use in bug pop-up infoboxes.

    This is meant to be embedded in other pages, rather than being
    a standalone page.
    """

    template_name = 'reviews/bug_infobox.html'

    HTML_ENTITY_RE = re.compile(r'(&[a-z]+;)')
    HTML_ENTITY_MAP = {
        '&quot;': '"',
        '&lt;': '<',
        '&gt;': '>',
        '&amp;': '&',
    }

    def get(
        self,
        request: HttpRequest,
        bug_id: str,
        **kwargs,
    ) -> HttpResponse:
        """Handle HTTP GET requests for this view.

        Args:
            request (django.http.HttpRequest):
                The HTTP request from the client.

            bug_id (str):
                The ID of the bug to view.

            *args (tuple):
                Positional arguments passed to the handler.

            **kwargs (dict):
                Keyword arguments passed to the handler.

        Returns:
            django.http.HttpResponse:
            The HTTP response to send to the client.

            If details on a bug could not be found or fetching bug information
            is not supported, this will return a a :http:`404`.
        """
        request = self.request
        review_request = self.review_request
        repository = review_request.repository

        if not repository:
            return HttpResponseNotFound(
                _('Review Request does not have an associated repository'))

        bug_tracker = repository.bug_tracker_service

        if not bug_tracker:
            return HttpResponseNotFound(
                _('Unable to find bug tracker service'))

        if not isinstance(bug_tracker, BaseBugTracker):
            return HttpResponseNotFound(
                _('Bug tracker %s does not support metadata')
                % bug_tracker.name)

        self.bug_id = bug_id
        self.bug_info = bug_tracker.get_bug_info(repository, bug_id)

        if (not self.bug_info.get('summary') and
            not self.bug_info.get('description')):
            return HttpResponseNotFound(
                _('No bug metadata found for bug %(bug_id)s on bug tracker '
                  '%(bug_tracker)s') % {
                    'bug_id': bug_id,
                    'bug_tracker': bug_tracker.name,
                })

        return super().get(request, **kwargs)

    def get_context_data(
        self,
        **kwargs,
    ) -> dict[str, Any]:
        """Return context data for the template.

        Args:
            **kwargs (dict):
                Keyword arguments passed to the view.

        Returns:
            dict:
            The resulting context data for the template.
        """
        description_text_format = self.bug_info.get('description_text_format',
                                                    'plain')
        description = self.normalize_text(self.bug_info['description'],
                                          description_text_format)

        bug_url = local_site_reverse(
            'bug_url',
            local_site=self.local_site,
            args=[self.review_request.display_id, self.bug_id])

        context_data = super().get_context_data(**kwargs)
        context_data.update({
            'bug_id': self.bug_id,
            'bug_url': bug_url,
            'bug_description': description,
            'bug_description_rich_text': description_text_format == 'markdown',
            'bug_status': self.bug_info['status'],
            'bug_summary': self.bug_info['summary'],
        })

        return context_data

    def normalize_text(
        self,
        text: str,
        text_format: str,
    ) -> SafeString:
        """Normalize the text for display.

        Based on the text format, this will sanitize and normalize the text
        so it's suitable for rendering to HTML.

        HTML text will have tags stripped away and certain common entities
        replaced.

        Markdown text will be rendered using our default Markdown parser
        rules.

        Plain text (or any unknown text format) will simply be escaped and
        wrapped, with paragraphs left intact.

        Args:
            text (str):
                The text to normalize for display.

            text_format (str):
                The text format. This should be one of ``html``, ``markdown``,
                or ``plain``.

        Returns:
            django.utils.safestring.SafeString:
            The resulting text, safe for rendering in HTML.
        """
        if text_format == 'html':
            # We want to strip the tags away, but keep certain common entities.
            text = (
                escape(self.HTML_ENTITY_RE.sub(
                    lambda m: (self.HTML_ENTITY_MAP.get(m.group(0)) or
                               m.group(0)),
                    strip_tags(text)))
                .replace('\n\n', '<br><br>'))
        elif text_format == 'markdown':
            # This might not know every bit of Markdown that's thrown at us,
            # but we'll do the best we can.
            text = render_markdown(text)
        else:
            # Should be plain text, but don't trust it.
            text = escape(text).replace('\n\n', '<br><br>')

        return mark_safe(text)


class BugURLRedirectView(ReviewRequestViewMixin, View):
    """Redirects the user to an external bug report."""

    def get(
        self,
        request: HttpRequest,
        bug_id: str,
        **kwargs,
    ) -> HttpResponse:
        """Handle HTTP GET requests for this view.

        Args:
            request (django.http.HttpRequest):
                The HTTP request from the client.

            bug_id (str):
                The ID of the bug report to redirect to.

            *args (tuple):
                Positional arguments passed to the handler.

            **kwargs (dict):
                Keyword arguments passed to the handler.

        Returns:
            django.http.HttpResponse:
            The HTTP response redirecting the client.
        """
        repository = self.review_request.repository

        if not repository:
            return HttpResponseNotFound(
                _('Review Request does not have an associated repository'))

        # Need to create a custom HttpResponse because a non-HTTP url scheme
        # will cause HttpResponseRedirect to fail with a "Disallowed Redirect".
        response = HttpResponse(status=302)
        response['Location'] = repository.bug_tracker % bug_id

        return response



"""A hook for adding information to the user infobox."""

from __future__ import annotations

from django.template.loader import render_to_string
from djblets.extensions.hooks import ExtensionHook, ExtensionHookPoint


class UserInfoboxHook(ExtensionHook, metaclass=ExtensionHookPoint):
    """A hook for adding information to the user infobox.

    Extensions can use this hook to add additional pieces of data to the box
    which pops up when hovering the mouse over a user.
    """

    def initialize(self, template_name=None):
        """Initialize the hook.

        Args:
            template_name (unicode):
                The template to render with the default :py:func:`render`
                method.
        """
        self.template_name = template_name

    def get_extra_context(self, user, request, local_site, **kwargs):
        """Return extra context to use when rendering the template.

        This may be overridden in order to make use of the default
        :py:func:`render` method.

        Args:
            user (django.contrib.auth.models.User):
                The user whose infobox is being shown.

            request (django.http.HttpRequest):
                The request for the infobox view.

            local_site (reviewboard.site.models.LocalSite):
                The local site, if any.

            **kwargs (dict):
                Additional keyword arguments.

        Returns:
            dict:
            Additional context to include when rendering the template.
        """
        return {}

    def get_etag_data(self, user, request, local_site, **kwargs):
        """Return data to be included in the user infobox ETag.

        The infobox view uses an ETag to enable browser caching of the content.
        If the extension returns data which can change, this method should
        return a string which is unique to that data.

        Args:
            user (django.contrib.auth.models.User):
                The user whose infobox is being shown.

            request (django.http.HttpRequest):
                The request for the infobox view.

            local_site (reviewboard.site.models.LocalSite):
                The local site, if any.

            **kwargs (dict):
                Additional keyword arguments.

        Returns:
            unicode:
            A string to be included in the ETag for the view.
        """
        return ''

    def render(self, user, request, local_site, **kwargs):
        """Return content to include in the user infobox.

        This may be overridden in the case where providing a custom template
        and overriding :py:func:`get_extra_context` is insufficient.

        Args:
            user (django.contrib.auth.models.User):
                The user whose infobox is being shown.

            request (django.http.HttpRequest):
                The request for the infobox view.

            local_site (reviewboard.site.models.LocalSite):
                The local site, if any.

            **kwargs (dict):
                Additional keyword arguments.

        Returns:
            django.utils.safestring.SafeText:
            Text to include in the infobox HTML.
        """
        assert self.template_name is not None

        context = {
            'extension': self.extension,
            'user': user,
        }
        context.update(self.get_extra_context(user=user,
                                              request=request,
                                              local_site=local_site))
        return render_to_string(
            template_name=self.template_name,
            context=context,
            request=request)



"""Mixins for review request views."""

from __future__ import annotations

import logging
from typing import Any, List, Optional, TYPE_CHECKING

from django.db.models import Q
from django.http import Http404, HttpRequest, HttpResponse
from django.shortcuts import get_object_or_404, render
from django.template.defaultfilters import date
from django.utils.formats import localize
from django.utils.html import format_html
from django.utils.safestring import SafeString, mark_safe
from django.utils.timezone import localtime
from django.utils.translation import gettext
from djblets.views.generic.base import (CheckRequestMethodViewMixin,
                                        PrePostDispatchViewMixin)
from typing_extensions import TypedDict

from reviewboard.accounts.mixins import CheckLoginRequiredViewMixin
from reviewboard.diffviewer.models import DiffSet
from reviewboard.reviews.models import ReviewRequest, ReviewRequestDraft
from reviewboard.site.mixins import CheckLocalSiteAccessViewMixin

if TYPE_CHECKING:
    from datetime import datetime

    from reviewboard.attachments.models import FileAttachmentSequence
    from reviewboard.reviews.models.base_review_request_details import \
        BaseReviewRequestDetails
    from reviewboard.reviews.models.review_request import \
        ReviewRequestCloseInfo
    from reviewboard.site.models import LocalSite

logger = logging.getLogger(__name__)


class ReviewRequestViewMixin(CheckRequestMethodViewMixin,
                             CheckLoginRequiredViewMixin,
                             CheckLocalSiteAccessViewMixin,
                             PrePostDispatchViewMixin):
    """Common functionality for all review request-related pages.

    This performs checks to ensure that the user has access to the page,
    returning an error page if not. It also provides common functionality
    for fetching a review request for the given page, returning suitable
    context for the template, and generating an image used to represent
    the site when posting to social media sites.
    """

    permission_denied_template_name = \
        'reviews/review_request_permission_denied.html'

    class StatusExtraInfo(TypedDict):
        """Extra info to include in the status rendering.

        This is used in :py:meth:`get_review_request_status_html`.
        """

        #: A text string to display to the user.
        #:
        #: Type:
        #:     str
        text: str

        #: The timestamp associated with this info block.
        #:
        #: Type:
        #:     datetime.datetime
        timestamp: Optional[datetime]

        #: Extra info to format into the ``text`` string.
        #:
        #: Type:
        #:     dict
        extra_vars: dict[str, Any]

    def pre_dispatch(
        self,
        request: HttpRequest,
        review_request_id: int,
        *args,
        **kwargs,
    ) -> Optional[HttpResponse]:
        """Look up objects and permissions before dispatching the request.

        This will first look up the review request, returning an error page
        if it's not accessible. It will then store the review request before
        calling the handler for the HTTP request.

        Args:
            request (django.http.HttpRequest):
                The HTTP request from the client.

            review_request_id (int):
                The ID of the review request being accessed.

            *args (tuple):
                Positional arguments to pass to the handler.

            **kwargs (dict):
                Keyword arguments to pass to the handler.

                These will be arguments provided by the URL pattern.

        Returns:
            django.http.HttpResponse:
            The resulting HTTP response to send to the client, if there's
            a Permission Denied.
        """
        self.review_request = self.get_review_request(
            review_request_id=review_request_id,
            local_site=self.local_site)

        if not self.review_request.is_accessible_by(request.user):
            return self.render_permission_denied(request)

        return None

    def render_permission_denied(
        self,
        request: HttpRequest,
    ) -> HttpResponse:
        """Render a Permission Denied page.

        This will be shown to the user if they're not able to view the
        review request.

        Args:
            request (django.http.HttpRequest):
                The HTTP request from the client.

        Returns:
            django.http.HttpResponse:
            The resulting HTTP response to send to the client.
        """
        return render(request,
                      self.permission_denied_template_name,
                      status=403)

    def get_review_request(
        self,
        review_request_id: int,
        local_site: Optional[LocalSite] = None,
    ) -> ReviewRequest:
        """Return the review request for the given display ID.

        Args:
            review_request_id (int):
                The review request's display ID.

            local_site (reviewboard.site.models.LocalSite):
                The Local Site the review request is on.

        Returns:
            reviewboard.reviews.models.review_request.ReviewRequest:
            The review request for the given display ID and Local Site.

        Raises:
            django.http.Http404:
                The review request could not be found.
        """
        q = ReviewRequest.objects.all()

        if local_site:
            q = q.filter(local_site=local_site,
                         local_id=review_request_id)
        else:
            q = q.filter(pk=review_request_id)

        q = q.select_related('submitter', 'repository')

        return get_object_or_404(q)

    def get_diff(
        self,
        revision: Optional[int] = None,
        draft: Optional[ReviewRequestDraft] = None,
    ) -> DiffSet:
        """Return a diff on the review request matching the given criteria.

        If a draft is provided, and ``revision`` is either ``None`` or matches
        the revision on the draft's DiffSet, that DiffSet will be returned.

        Args:
            revision (int, optional):
                The revision of the diff to retrieve. If not provided, the
                latest DiffSet will be returned.

            draft (reviewboard.reviews.models.review_request_draft.
                   ReviewRequestDraft, optional):
                The draft of the review request.

        Returns:
            reviewboard.diffviewer.models.diffset.DiffSet:
            The resulting DiffSet.

        Raises:
            django.http.Http404:
                The diff does not exist.
        """
        # Normalize the revision, since it might come in as a string.
        if revision:
            revision = int(revision)

        # This will try to grab the diff associated with a draft if the review
        # request has an associated draft and is either the revision being
        # requested or no revision is being requested.
        if (draft and draft.diffset_id and
            (revision is None or draft.diffset.revision == revision)):
            return draft.diffset

        query = Q(history=self.review_request.diffset_history_id)

        # Grab a revision if requested.
        if revision is not None:
            query = query & Q(revision=revision)

        try:
            return DiffSet.objects.filter(query).latest()
        except DiffSet.DoesNotExist:
            raise Http404

    def get_social_page_image_url(
        self,
        file_attachments: FileAttachmentSequence,
    ) -> Optional[str]:
        """Return the URL to an image used for social media sharing.

        This will look for the first attachment in a list of attachments that
        can be used to represent the review request on social media sites and
        chat services. If a suitable attachment is found, its URL will be
        returned.

        Args:
            file_attachments (list of reviewboard.attachments.models.
                              FileAttachment):
                A list of file attachments used on a review request.

        Returns:
            str:
            The URL to the first image file attachment, if found, or ``None``
            if no suitable attachments were found.
        """
        for file_attachment in file_attachments:
            if file_attachment.mimetype.startswith('image/'):
                return file_attachment.get_absolute_url()

        return None

    def get_review_request_status_html(
        self,
        review_request_details: BaseReviewRequestDetails,
        close_info: ReviewRequestCloseInfo,
        extra_info: List[ReviewRequestViewMixin.StatusExtraInfo] = [],
    ) -> SafeString:
        """Return HTML describing the current status of a review request.

        This will return a description of the submitted, discarded, or open
        state for the review request, for use in the rendering of the page.

        Args:
            review_request_details (reviewboard.reviews.models
                                    .base_review_request_details
                                    .BaseReviewRequestDetails):
                The review request or draft being viewed.

            close_info (dict):
                A dictionary of information on the closed state of the
                review request.

            extra_info (list of dict):
                A list of dictionaries showing additional status information.
                Each must have a ``text`` field containing a format string
                using ``{keyword}``-formatted variables, a ``timestamp`` field
                (which will be normalized to the local timestamp), and an
                optional ``extra_vars`` for the format string.

        Returns:
            django.utils.safestring.SafeString:
            The status text as HTML for the page.
        """
        review_request = self.review_request
        status = review_request.status

        if status == ReviewRequest.SUBMITTED:
            timestamp = close_info['timestamp']

            if timestamp:
                text = gettext('Created {created_time} and submitted '
                               '{timestamp}')
            else:
                text = gettext('Created {created_time} and submitted')
        elif status == ReviewRequest.DISCARDED:
            timestamp = close_info['timestamp']

            if timestamp:
                text = gettext('Created {created_time} and discarded '
                               '{timestamp}')
            else:
                text = gettext('Created {created_time} and discarded')
        elif status == ReviewRequest.PENDING_REVIEW:
            text = gettext('Created {created_time} and updated {timestamp}')
            timestamp = review_request_details.last_updated
        else:
            logger.error('Unexpected review request status %r for '
                         'review request %s',
                         status, review_request.display_id)

            return mark_safe('')

        parts = [
            {
                'text': text,
                'timestamp': timestamp,
                'extra_vars': {
                    'created_time': date(localtime(review_request.time_added)),
                },
            },
        ] + extra_info

        html_parts = []

        for part in parts:
            if part['timestamp']:
                timestamp = localtime(part['timestamp'])
                timestamp_html = format_html(
                    '<time class="timesince" datetime="{0}">{1}</time>',
                    timestamp.isoformat(),
                    localize(timestamp))
            else:
                timestamp_html = ''

            html_parts.append(format_html(
                part['text'],
                timestamp=timestamp_html,
                **part.get('extra_vars', {})))

        return mark_safe(' &mdash; '.join(html_parts))



"""Unit tests for reviewboard.site.mixins.CheckLocalSiteAccessViewMixin."""

from django.contrib.auth.models import User
from django.http import HttpResponse, HttpResponseRedirect
from django.views.generic.base import View

from reviewboard.site.mixins import CheckLocalSiteAccessViewMixin
from reviewboard.testing.testcase import TestCase


class CheckLocalSiteAccessViewMixinTests(TestCase):
    """Unit tests for CheckLocalSiteAccessViewMixin."""

    fixtures = ['test_users', 'test_site']

    def test_dispatch_with_local_site_and_allowed(self):
        """Testing CheckLocalSiteAccessViewMixin.dispatch with LocalSite and
        access allowed
        """
        class MyView(CheckLocalSiteAccessViewMixin, View):
            def get(view, *args, **kwargs):
                self.assertIsNotNone(view.local_site)
                self.assertEqual(view.local_site.name, 'local-site-1')

                return HttpResponse('success')

        local_site = self.get_local_site(self.local_site_name)
        request = self.create_http_request(user=local_site.users.all()[0],
                                           local_site=local_site)

        view = MyView.as_view()
        response = view(request, local_site_name=local_site.name)

        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.content, b'success')

    def test_dispatch_with_local_site_and_not_allowed(self):
        """Testing CheckLocalSiteAccessViewMixin.dispatch with LocalSite and
        access not allowed
        """
        class MyView(CheckLocalSiteAccessViewMixin, View):
            def get(view, *args, **kwargs):
                self.assertIsNotNone(view.local_site)
                self.assertEqual(view.local_site.name, 'local-site-1')

                return HttpResponse('success')

        view = MyView.as_view()

        local_site = self.get_local_site(self.local_site_name)
        request = self.create_http_request(
            user=User.objects.create_user(username='test123',
                                          email='test123@example.com'),
            local_site=local_site,
            view=view)

        response = view(request, local_site_name=local_site.name)
        self.assertEqual(response.status_code, 403)

    def test_dispatch_with_local_site_and_anonymous(self):
        """Testing CheckLocalSiteAccessViewMixin.dispatch with LocalSite and
        anonymous user
        """
        class MyView(CheckLocalSiteAccessViewMixin, View):
            def get(view, *args, **kwargs):
                self.assertIsNotNone(view.local_site)
                self.assertEqual(view.local_site.name, 'local-site-1')

                return HttpResponse('success')

        view = MyView.as_view()

        local_site = self.get_local_site(self.local_site_name)
        request = self.create_http_request(local_site=local_site,
                                           view=view)

        response = view(request, local_site_name=local_site.name)
        self.assertIsInstance(response, HttpResponseRedirect)

    def test_dispatch_with_no_local_site(self):
        """Testing CheckLocalSiteAccessViewMixin.dispatch with no LocalSite"""
        class MyView(CheckLocalSiteAccessViewMixin, View):
            def get(view, *args, **kwargs):
                self.assertIsNone(view.local_site)

                return HttpResponse('success')

        view = MyView.as_view()

        request = self.create_http_request(
            user=User.objects.get(username='doc'),
            view=view)

        response = view(request)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.content, b'success')
